"""Resolve legacy peer DIDs.

Resolution is performed by looking up a stored DID Document.
"""

import logging
from copy import deepcopy
from dataclasses import asdict, dataclass
from typing import List, Optional, Sequence, Text, Union

from pydid import DID

from ...cache.base import BaseCache
from ...config.injection_context import InjectionContext
from ...connections.base_manager import BaseConnectionManager
from ...core.profile import Profile
from ...did.did_key import DIDKey
from ...messaging.valid import IndyDID
from ...storage.error import StorageNotFoundError
from ...wallet.key_type import ED25519
from ..base import BaseDIDResolver, DIDNotFound, ResolverType

LOGGER = logging.getLogger(__name__)


class LegacyDocCorrections:
    """Legacy peer DID document corrections.

    These corrections align the document with updated DID spec and DIDComm
    conventions. This also helps with consistent processing of DID Docs.

    Input example:

    ::

        {
            "@context": "https://w3id.org/did/v1",
            "id": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ",
            "publicKey": [
                {
                "id": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ#1",
                "type": "Ed25519VerificationKey2018",
                "controller": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ",
                "publicKeyBase58": "AU2FFjtkVzjFuirgWieqGGqtNrAZWS9LDuB8TDp6EUrG"
                }
            ],
            "authentication": [
                {
                "type": "Ed25519SignatureAuthentication2018",
                "publicKey": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ#1"
                }
            ],
            "service": [
                {
                "id": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ;indy",
                "type": "IndyAgent",
                "priority": 0,
                "recipientKeys": [
                    "AU2FFjtkVzjFuirgWieqGGqtNrAZWS9LDuB8TDp6EUrG"
                ],
                "routingKeys": ["9NnKFUZoYcCqYC2PcaXH3cnaGsoRfyGgyEHbvbLJYh8j"],
                "serviceEndpoint": "http://bob:3000"
                }
            ]
        }

    Output example:

    ::

        {
            "@context": "https://w3id.org/did/v1",
            "id": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ",
            "verificationMethod": [
                {
                "id": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ#1",
                "type": "Ed25519VerificationKey2018",
                "controller": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ",
                "publicKeyBase58": "AU2FFjtkVzjFuirgWieqGGqtNrAZWS9LDuB8TDp6EUrG"
                }
            ],
            "authentication": ["did:sov:JNKL9kJxQi5pNCfA8QBXdJ#1"],
            "service": [
                {
                "id": "did:sov:JNKL9kJxQi5pNCfA8QBXdJ#didcomm",
                "type": "did-communication",
                "priority": 0,
                "recipientKeys": ["did:sov:JNKL9kJxQi5pNCfA8QBXdJ#1"],
                "routingKeys": [
                    "did:key:z6Mknq3MqipEt9hJegs6J9V7tiLa6T5H5rX3fFCXksJKTuv7#z6Mknq3MqipEt9hJegs6J9V7tiLa6T5H5rX3fFCXksJKTuv7"
                ],
                "serviceEndpoint": "http://bob:3000"
                }
            ]
        }

    """

    @staticmethod
    def public_key_is_verification_method(value: dict) -> dict:
        """Replace publicKey with verificationMethod."""
        if "publicKey" in value:
            value["verificationMethod"] = value.pop("publicKey")
        return value

    @staticmethod
    def authentication_is_list_of_verification_methods_and_refs(value: dict) -> dict:
        """Update authentication to be a list of methods and references."""
        if "authentication" in value:
            modified = []
            for authn in value["authentication"]:
                if isinstance(authn, dict) and "publicKey" in authn:
                    modified.append(authn["publicKey"])
                else:
                    modified.append(authn)
                # TODO more checks?
            value["authentication"] = modified
        return value

    @staticmethod
    def didcomm_services_use_updated_conventions(value: dict) -> dict:
        """Update DIDComm services to use updated conventions."""
        if "service" in value:
            for index, service in enumerate(value["service"]):
                if "type" in service and service["type"] == "IndyAgent":
                    service["type"] = "did-communication"
                    if ";" in service["id"]:
                        service["id"] = value["id"] + f"#didcomm-{index}"
                    if "#" not in service["id"]:
                        service["id"] += f"#didcomm-{index}"
                    if "priority" in service and service["priority"] is None:
                        service.pop("priority")
        return value

    @staticmethod
    def recip_base58_to_ref(vms: List[dict], recip: str) -> str:
        """Convert base58 public key to ref."""
        for vm in vms:
            if "publicKeyBase58" in vm and vm["publicKeyBase58"] == recip:
                return vm["id"]
        return recip

    @classmethod
    def did_key_to_did_key_ref(cls, key: str):
        """Convert did:key to did:key ref."""
        # Check if key is already a ref
        if key.rfind("#") != -1:
            return key
        # Get the value after removing did:key:
        value = key.replace("did:key:", "")

        return key + "#" + value

    @classmethod
    def didcomm_services_recip_keys_are_refs_routing_keys_are_did_key_ref(
        cls,
        value: dict,
    ) -> dict:
        """Update DIDComm service recips to use refs and routingKeys to use did:key."""
        vms = value.get("verificationMethod", [])
        if "service" in value:
            for service in value["service"]:
                if "type" in service and service["type"] == "did-communication":
                    service["recipientKeys"] = [
                        cls.recip_base58_to_ref(vms, recip)
                        for recip in service.get("recipientKeys", [])
                    ]
                if "routingKeys" in service:
                    service["routingKeys"] = [
                        (
                            DIDKey.from_public_key_b58(key, ED25519).key_id
                            if "did:key:" not in key
                            else cls.did_key_to_did_key_ref(key)
                        )
                        for key in service["routingKeys"]
                    ]
        return value

    @staticmethod
    def qualified(did_or_did_url: str) -> str:
        """Make sure DID or DID URL is fully qualified."""
        if not did_or_did_url.startswith("did:"):
            return f"did:sov:{did_or_did_url}"
        return did_or_did_url

    @classmethod
    def fully_qualified_ids_and_controllers(cls, value: dict) -> dict:
        """Make sure IDs and controllers are fully qualified."""

        def _make_qualified(value: dict) -> dict:
            if "id" in value:
                ident = value["id"]
                value["id"] = cls.qualified(ident)
            if "controller" in value:
                controller = value["controller"]
                value["controller"] = cls.qualified(controller)
            return value

        value = _make_qualified(value)
        vms = []
        for verification_method in value.get("verificationMethod", []):
            vms.append(_make_qualified(verification_method))

        services = []
        for service in value.get("service", []):
            services.append(_make_qualified(service))

        auths = []
        for authn in value.get("authentication", []):
            if isinstance(authn, dict):
                auths.append(_make_qualified(authn))
            elif isinstance(authn, str):
                auths.append(cls.qualified(authn))
            else:
                raise ValueError("Unexpected authentication value type")

        value["authentication"] = auths
        value["verificationMethod"] = vms
        value["service"] = services
        return value

    @staticmethod
    def remove_verification_method(vms: List[dict], public_key_base58: str) -> List[dict]:
        """Remove the verification method with the given key."""
        return [vm for vm in vms if vm["publicKeyBase58"] != public_key_base58]

    @classmethod
    def remove_routing_keys_from_verification_method(cls, value: dict) -> dict:
        """Remove routing keys from verification methods.

        This was an old convention; routing keys were added to the public keys
        of the doc even though they're usually not owned by the doc sender.

        This correction should be applied before turning the routing keys into
        did keys.
        """
        vms = value.get("verificationMethod", [])
        for service in value.get("service", []):
            if "routingKeys" in service:
                for routing_key in service["routingKeys"]:
                    vms = cls.remove_verification_method(vms, routing_key)
        value["verificationMethod"] = vms
        return value

    @classmethod
    def apply(cls, value: dict) -> dict:
        """Apply all corrections to the given DID document."""
        value = deepcopy(value)
        for correction in (
            cls.public_key_is_verification_method,
            cls.authentication_is_list_of_verification_methods_and_refs,
            cls.fully_qualified_ids_and_controllers,
            cls.didcomm_services_use_updated_conventions,
            cls.remove_routing_keys_from_verification_method,
            cls.didcomm_services_recip_keys_are_refs_routing_keys_are_did_key_ref,
        ):
            value = correction(value)

        return value


@dataclass
class RetrieveResult:
    """Entry in the peer DID cache."""

    is_local: bool
    doc: Optional[dict] = None


class LegacyPeerDIDResolver(BaseDIDResolver):
    """Resolve legacy peer DIDs."""

    def __init__(self):
        """Initialize the resolver instance."""
        super().__init__(ResolverType.NATIVE)

    async def setup(self, context: InjectionContext):
        """Perform required setup for the resolver."""

    async def _fetch_did_document(self, profile: Profile, did: str) -> RetrieveResult:
        """Fetch DID from wallet if available.

        This is the method to be used with fetch_did_document to enable caching.
        """
        conn_mgr = BaseConnectionManager(profile)
        if did.startswith("did:sov:"):
            did = did[8:]
        try:
            doc, _ = await conn_mgr.fetch_did_document(did)
            LOGGER.debug("Fetched doc %s", doc)
            to_cache = RetrieveResult(True, doc=doc)
        except StorageNotFoundError:
            LOGGER.debug("Failed to fetch doc for did %s", did)
            to_cache = RetrieveResult(False)

        return to_cache

    async def fetch_did_document(
        self, profile: Profile, did: str, *, ttl: Optional[int] = None
    ):
        """Fetch DID from wallet if available.

        Return value is cached.
        """
        cache_key = f"resolver::LegacyPeerDIDResolver::{did}"
        cache = profile.inject_or(BaseCache)
        if cache:
            async with cache.acquire(cache_key) as entry:
                if entry.result:
                    result = RetrieveResult(**entry.result)
                else:
                    result = await self._fetch_did_document(profile, did)
                    await entry.set_result(asdict(result), ttl or self.DEFAULT_TTL)
        else:
            result = await self._fetch_did_document(profile, did)

        return result

    async def supports(self, profile: Profile, did: str) -> bool:
        """Return whether this resolver supports the given DID.

        This resolver resolves unqualified DIDs and dids prefixed with
        `did:sov:`.

        These DIDs have the unfortunate characteristic of overlapping with what
        ACA-Py uses for DIDs written to Indy ledgers. This means that we will
        need to attempt to resolve but defer to the Indy resolver if we don't
        find anything locally. This has the side effect that this resolver will
        never raise DIDNotFound since it won't even be selected for resolution
        unless we have the DID in our wallet.

        This will check if the DID matches the IndyDID regex. If it does,
        attempt a lookup in the wallet for a document. If found, return True.
        Else, return False.
        """
        LOGGER.debug("Checking if resolver supports DID %s", did)
        if IndyDID.PATTERN.match(did):
            LOGGER.debug("DID is valid IndyDID %s", did)
            result = await self.fetch_did_document(profile, did)
            return result.is_local
        else:
            return False

    async def resolve(
        self,
        profile: Profile,
        did: Union[str, DID],
        service_accept: Optional[Sequence[Text]] = None,
    ) -> dict:
        """Resolve a Legacy Peer DID to a DID document by fetching from the wallet.

        This overrides the default resolve method so we can take care of caching
        ourselves since we use it for the supports method as well.
        """
        return await self._resolve(profile, str(did), service_accept)

    async def _resolve(
        self,
        profile: Profile,
        did: str,
        service_accept: Optional[Sequence[Text]] = None,
    ) -> dict:
        """Resolve Legacy Peer DID to a DID document by fetching from the wallet.

        By the time this resolver is selected, it should be impossible for it
        to raise a DIDNotFound.
        """
        result = await self.fetch_did_document(profile, did)
        if result.is_local:
            assert result.doc
            return LegacyDocCorrections.apply(result.doc)
        else:
            # This should be impossible because of the checks in supports
            raise DIDNotFound(f"DID not found: {did}")



"""Indy-VDR ledger implementation."""

import asyncio
import hashlib
import json
import logging
import os
import os.path
import tempfile
from datetime import date, datetime, timezone
from io import StringIO
from pathlib import Path
from time import time
from typing import List, Optional, Tuple, Union

from indy_vdr import Pool, Request, VdrError, ledger, open_pool

from ..cache.base import BaseCache
from ..core.profile import Profile
from ..messaging.valid import IndyDID
from ..storage.base import BaseStorage, StorageRecord
from ..utils import sentinel
from ..utils.env import storage_path
from ..utils.general import strip_did_prefix
from ..wallet.base import BaseWallet, DIDInfo
from ..wallet.did_posture import DIDPosture
from ..wallet.error import WalletNotFoundError
from .base import BaseLedger, Role
from .endpoint_type import EndpointType
from .error import (
    BadLedgerRequestError,
    ClosedPoolError,
    LedgerConfigError,
    LedgerError,
    LedgerTransactionError,
)
from .util import TAA_ACCEPTED_RECORD_TYPE

LOGGER = logging.getLogger(__name__)


def _normalize_txns(txns: str) -> str:
    """Normalize a set of genesis transactions."""
    lines = StringIO()
    for line in txns.splitlines():
        line = line.strip()
        if line:
            lines.write(line)
            lines.write("\n")
    return lines.getvalue()


def _write_safe(path: Path, content: str):
    """Atomically write to a file path."""
    dir_path = path.parent
    with tempfile.NamedTemporaryFile(dir=dir_path, delete=False) as tmp:
        tmp.write(content.encode("utf-8"))
        tmp_name = tmp.name
    os.rename(tmp_name, path)


def _hash_txns(txns: str) -> str:
    """Obtain a hash of a set of genesis transactions."""
    return hashlib.sha256(txns.encode("utf-8")).hexdigest()[-16:]


class IndyVdrLedgerPool:
    """Indy-VDR ledger pool manager."""

    def __init__(
        self,
        name: str,
        *,
        keepalive: int = 0,
        cache: Optional[BaseCache] = None,
        cache_duration: int = 600,
        genesis_transactions: Optional[str] = None,
        read_only: bool = False,
        socks_proxy: Optional[str] = None,
    ):
        """Initialize an IndyLedger instance.

        Args:
            name: The pool ledger configuration name
            keepalive: How many seconds to keep the ledger open
            cache: The cache instance to use
            cache_duration: The TTL for ledger cache entries
            genesis_transactions: The ledger genesis transaction as a string
            read_only: Prevent any ledger write operations
            socks_proxy: Specifies socks proxy for ZMQ to connect to ledger pool
        """
        self.ref_count = 0
        self.ref_lock = asyncio.Lock()
        self.keepalive = keepalive
        self.close_task: asyncio.Future = None
        self.cache = cache
        self.cache_duration: int = cache_duration
        self.handle: Optional[Pool] = None
        self.name = name
        self.cfg_path_cache: Optional[Path] = None
        self.genesis_hash_cache: Optional[str] = None
        self.genesis_txns_cache = genesis_transactions
        self.init_config = bool(genesis_transactions)
        self.taa_cache: Optional[dict] = None
        self.read_only: bool = read_only
        self.socks_proxy: str = socks_proxy

    @property
    def cfg_path(self) -> Path:
        """Get the path to the configuration file, ensuring it's created."""
        if not self.cfg_path_cache:
            self.cfg_path_cache = storage_path("vdr", create=True)
        return self.cfg_path_cache

    @property
    def genesis_hash(self) -> str:
        """Get the hash of the configured genesis transactions."""
        if not self.genesis_hash_cache:
            self.genesis_hash_cache = _hash_txns(self.genesis_txns)
        return self.genesis_hash_cache

    @property
    def genesis_txns(self) -> str:
        """Get the configured genesis transactions."""
        if not self.genesis_txns_cache:
            try:
                path = self.cfg_path.joinpath(self.name, "genesis")
                self.genesis_txns_cache = _normalize_txns(open(path).read())
            except FileNotFoundError:
                raise LedgerConfigError("Pool config '%s' not found", self.name) from None
        return self.genesis_txns_cache

    async def create_pool_config(self, genesis_transactions: str, recreate: bool = False):
        """Create the pool ledger configuration."""

        cfg_pool = self.cfg_path.joinpath(self.name)
        cfg_pool.mkdir(exist_ok=True)
        genesis = _normalize_txns(genesis_transactions)
        if not genesis:
            raise LedgerConfigError("Empty genesis transactions")

        genesis_path = cfg_pool.joinpath("genesis")
        try:
            cmp_genesis = open(genesis_path).read()
            if _normalize_txns(cmp_genesis) == genesis:
                LOGGER.debug(
                    "Pool ledger config '%s' is consistent, skipping write",
                    self.name,
                )
                return
            elif not recreate:
                raise LedgerConfigError(
                    f"Pool ledger '{self.name}' exists with "
                    "different genesis transactions"
                )
        except FileNotFoundError:
            pass

        try:
            _write_safe(genesis_path, genesis)
        except OSError as err:
            raise LedgerConfigError("Error writing genesis transactions") from err
        LOGGER.debug("Wrote pool ledger config '%s'", self.name)

        self.genesis_txns_cache = genesis

    async def open(self):
        """Open the pool ledger, creating it if necessary."""

        if self.init_config:
            await self.create_pool_config(self.genesis_txns_cache, recreate=True)
            self.init_config = False

        genesis_hash = self.genesis_hash
        cfg_pool = self.cfg_path.joinpath(self.name)
        cfg_pool.mkdir(exist_ok=True)

        cache_path = cfg_pool.joinpath(f"cache-{genesis_hash}")
        try:
            txns = open(cache_path).read()
            cached = True
        except FileNotFoundError:
            txns = self.genesis_txns
            cached = False

        self.handle = await open_pool(transactions=txns, socks_proxy=self.socks_proxy)
        upd_txns = _normalize_txns(await self.handle.get_transactions())
        if not cached or upd_txns != txns:
            try:
                _write_safe(cache_path, upd_txns)
            except OSError:
                LOGGER.exception("Error writing cached genesis transactions")

    async def close(self):
        """Close the pool ledger."""
        if self.handle:
            exc = None
            for attempt in range(3):
                try:
                    self.handle.close()
                except VdrError as err:
                    await asyncio.sleep(0.01)
                    exc = err
                    continue

                self.handle = None
                exc = None
                break

            if exc:
                LOGGER.exception("Exception when closing pool ledger", exc_info=exc)
                self.ref_count += 1  # if we are here, we should have self.ref_lock
                self.close_task = None
                raise LedgerError("Exception when closing pool ledger") from exc

    async def context_open(self):
        """Open the ledger if necessary and increase the number of active references."""
        async with self.ref_lock:
            if self.close_task:
                self.close_task.cancel()
            if not self.handle:
                LOGGER.debug("Opening the pool ledger")
                await self.open()
            self.ref_count += 1

    async def context_close(self):
        """Release the reference and schedule closing of the pool ledger."""

        async def closer(timeout: int):
            """Close the pool ledger after a timeout."""
            await asyncio.sleep(timeout)
            async with self.ref_lock:
                if not self.ref_count:
                    LOGGER.debug("Closing pool ledger after timeout")
                    await self.close()

        async with self.ref_lock:
            self.ref_count -= 1
            if not self.ref_count:
                if self.keepalive:
                    self.close_task = asyncio.ensure_future(closer(self.keepalive))
                else:
                    await self.close()


class IndyVdrLedger(BaseLedger):
    """Indy-VDR ledger class."""

    BACKEND_NAME = "indy-vdr"

    def __init__(
        self,
        pool: IndyVdrLedgerPool,
        profile: Profile,
    ):
        """Initialize an IndyVdrLedger instance.

        Args:
            pool: The pool instance handling the raw ledger connection
            profile: The active profile instance
        """
        self.pool = pool
        self.profile = profile

    @property
    def pool_handle(self):
        """Accessor for the ledger pool handle."""
        return self.pool.handle

    @property
    def pool_name(self) -> str:
        """Accessor for the ledger pool name."""
        return self.pool.name

    @property
    def read_only(self) -> bool:
        """Accessor for the ledger read-only flag."""
        return self.pool.read_only

    async def is_ledger_read_only(self) -> bool:
        """Check if ledger is read-only including TAA."""
        if self.read_only:
            return self.read_only
        # if TAA is required and not accepted we should be in read-only mode
        taa = await self.get_txn_author_agreement()
        if taa["taa_required"]:
            taa_acceptance = await self.get_latest_txn_author_acceptance()
            if "mechanism" not in taa_acceptance:
                return True
        return self.read_only

    async def __aenter__(self) -> "IndyVdrLedger":
        """Context manager entry.

        Returns:
            The current instance

        """
        await super().__aenter__()
        await self.pool.context_open()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        """Context manager exit."""
        await self.pool.context_close()
        await super().__aexit__(exc_type, exc, tb)

    async def _submit(
        self,
        request: Union[str, Request],
        sign: Optional[bool] = None,
        taa_accept: Optional[bool] = None,
        sign_did: DIDInfo = sentinel,
        write_ledger: bool = True,
    ) -> dict:
        """Sign and submit request to ledger.

        Args:
            request: The request to submit
            sign: whether or not to sign the request
            taa_accept: whether to apply TAA acceptance to the (signed, write) request
            sign_did: override the signing DID
            write_ledger: whether to write the request to the ledger

        """

        if not self.pool_handle:
            raise ClosedPoolError(
                f"Cannot sign and submit request to closed pool '{self.pool_name}'"
            )

        if isinstance(request, str):
            request = ledger.build_custom_request(request)
        elif not isinstance(request, Request):
            raise BadLedgerRequestError("Expected str or Request")

        if sign is None or sign:
            if sign_did is sentinel:
                sign_did = await self.get_wallet_public_did()
            if sign is None:
                sign = bool(sign_did)

        if sign:
            if not sign_did:
                raise BadLedgerRequestError("Cannot sign request without a public DID")

            if taa_accept or taa_accept is None:
                acceptance = await self.get_latest_txn_author_acceptance()
                if acceptance:
                    acceptance = {
                        "taaDigest": acceptance["digest"],
                        "mechanism": acceptance["mechanism"],
                        "time": acceptance["time"],
                    }
                    request.set_txn_author_agreement_acceptance(acceptance)

            async with self.profile.session() as session:
                wallet = session.inject(BaseWallet)
                request.set_signature(
                    await wallet.sign_message(request.signature_input, sign_did.verkey)
                )
                del wallet

        if not write_ledger:
            return json.loads(request.body)

        try:
            request_result = await self.pool.handle.submit_request(request)
        except VdrError as err:
            raise LedgerTransactionError("Ledger request error") from err

        return request_result

    async def _create_schema_request(
        self,
        public_info: DIDInfo,
        schema_json: str,
        write_ledger: bool = True,
        endorser_did: Optional[str] = None,
    ):
        """Create the ledger request for publishing a schema."""
        try:
            schema_req = ledger.build_schema_request(
                strip_did_prefix(public_info.did), schema_json
            )
        except VdrError as err:
            raise LedgerError("Exception when building schema request") from err

        if endorser_did and not write_ledger:
            schema_req.set_endorser(endorser_did)

        return schema_req

    async def _create_revoc_reg_def_request(
        self,
        public_info: DIDInfo,
        revoc_reg_def_json: str,
        write_ledger: bool = True,
        endorser_did: Optional[str] = None,
    ):
        """Create the ledger request for publishing a revocation registry definition."""
        try:
            revoc_reg_def_req = ledger.build_revoc_reg_def_request(
                public_info.did, revoc_reg_def_json
            )
        except VdrError as err:
            raise LedgerError("Exception when building revoc reg def request") from err

        if endorser_did and not write_ledger:
            revoc_reg_def_req.set_endorser(endorser_did)

        return revoc_reg_def_req

    async def _create_revoc_reg_entry_request(
        self,
        public_info: DIDInfo,
        revoc_reg_def_id: str,
        revoc_def_type: str,
        revoc_reg_entry_json: str,
        write_ledger: bool = True,
        endorser_did: Optional[str] = None,
    ):
        """Create the ledger request for publishing a revocation registry definition."""
        try:
            revoc_reg_entry_request = ledger.build_revoc_reg_entry_request(
                public_info.did, revoc_reg_def_id, revoc_def_type, revoc_reg_entry_json
            )
        except VdrError as err:
            raise LedgerError("Exception when building revoc reg def request") from err

        if endorser_did and not write_ledger:
            revoc_reg_entry_request.set_endorser(endorser_did)

        return revoc_reg_entry_request

    async def get_schema(self, schema_id: str) -> dict:
        """Get a schema from the cache if available, otherwise fetch from the ledger.

        Args:
            schema_id: The schema id (or stringified sequence number) to retrieve

        """
        if self.pool.cache:
            result = await self.pool.cache.get(f"schema::{schema_id}")
            if result:
                return result

        if schema_id.isdigit():
            return await self.fetch_schema_by_seq_no(int(schema_id))
        else:
            return await self.fetch_schema_by_id(schema_id)

    async def fetch_schema_by_id(self, schema_id: str) -> dict:
        """Get schema from ledger.

        Args:
            schema_id: The schema id (or stringified sequence number) to retrieve

        Returns:
            Indy schema dict

        """

        public_info = await self.get_wallet_public_did()
        public_did = public_info.did if public_info else None

        try:
            schema_req = ledger.build_get_schema_request(
                strip_did_prefix(public_did), strip_did_prefix(schema_id)
            )
        except VdrError as err:
            raise LedgerError("Exception when building get-schema request") from err

        response = await self._submit(schema_req, sign_did=public_info)

        schema_seqno = response.get("seqNo")
        if not schema_seqno:
            return None  # schema not found

        schema_name = response["data"]["name"]
        schema_version = response["data"]["version"]
        schema_id = f"{response['dest']}:2:{schema_name}:{schema_version}"
        schema_data = {
            "ver": "1.0",
            "id": schema_id,
            "name": schema_name,
            "version": schema_version,
            "attrNames": response["data"]["attr_names"],
            "seqNo": schema_seqno,
        }

        if self.pool.cache:
            await self.pool.cache.set(
                [f"schema::{schema_id}", f"schema::{schema_seqno}"],
                schema_data,
                self.pool.cache_duration,
            )

        return schema_data

    async def fetch_schema_by_seq_no(self, seq_no: int) -> dict:
        """Fetch a schema by its sequence number.

        Args:
            seq_no: schema ledger sequence number

        Returns:
            Indy schema dict

        """
        # get txn by sequence number, retrieve schema identifier components
        try:
            request = ledger.build_get_txn_request(None, None, seq_no=seq_no)
        except VdrError as err:
            raise LedgerError("Exception when building get-txn request") from err

        response = await self._submit(request)

        # transaction data format assumes node protocol >= 1.4 (circa 2018-07)
        data_txn = (response.get("data", {}) or {}).get("txn", {})
        if data_txn.get("type", None) == "101":  # marks indy-sdk schema txn type
            (origin_did, name, version) = (
                data_txn["metadata"]["from"],
                data_txn["data"]["data"]["name"],
                data_txn["data"]["data"]["version"],
            )
            schema_id = f"{origin_did}:2:{name}:{version}"
            return await self.get_schema(schema_id)

        raise LedgerTransactionError(
            f"Could not get schema from ledger for seq no {seq_no}"
        )

    async def _create_credential_definition_request(
        self,
        public_info: DIDInfo,
        credential_definition_json: str,
        write_ledger: bool = True,
        endorser_did: Optional[str] = None,
    ):
        """Create the ledger request for publishing a credential definition."""
        try:
            cred_def_req = ledger.build_cred_def_request(
                public_info.did, credential_definition_json
            )
        except VdrError as err:
            raise LedgerError("Exception when building cred def request") from err

        if endorser_did and not write_ledger:
            cred_def_req.set_endorser(endorser_did)

        return cred_def_req

    async def get_credential_definition(self, credential_definition_id: str) -> dict:
        """Get a credential definition from the cache if available, otherwise the ledger.

        Args:
            credential_definition_id: The schema id of the schema to fetch cred def for

        """
        if self.pool.cache:
            cache_key = f"credential_definition::{credential_definition_id}"
            async with self.pool.cache.acquire(cache_key) as entry:
                if entry.result:
                    result = entry.result
                else:
                    result = await self.fetch_credential_definition(
                        credential_definition_id
                    )
                    if result:
                        await entry.set_result(result, self.pool.cache_duration)
                return result

        return await self.fetch_credential_definition(credential_definition_id)

    async def fetch_credential_definition(
        self, credential_definition_id: str
    ) -> dict | None:
        """Get a credential definition from the ledger by id.

        Args:
            credential_definition_id: The cred def id of the cred def to fetch

        """

        public_info = await self.get_wallet_public_did()
        public_did = public_info.did if public_info else None

        try:
            cred_def_req = ledger.build_get_cred_def_request(
                strip_did_prefix(public_did), strip_did_prefix(credential_definition_id)
            )
        except VdrError as err:
            raise LedgerError("Exception when building get-cred-def request") from err

        response = await self._submit(cred_def_req, sign_did=public_info)
        if not response["data"]:
            return None

        schema_id = str(response["ref"])
        signature_type = response["signature_type"]
        tag = response.get("tag", "default")
        origin_did = response["origin"]
        # FIXME: issuer has a method to create a cred def ID
        # may need to qualify the DID
        cred_def_id = f"{origin_did}:3:{signature_type}:{schema_id}:{tag}"

        return {
            "ver": "1.0",
            "id": cred_def_id,
            "schemaId": schema_id,
            "type": signature_type,
            "tag": tag,
            "value": response["data"],
        }

    async def credential_definition_id2schema_id(self, credential_definition_id):
        """From a credential definition, get the identifier for its schema.

        Args:
            credential_definition_id: The identifier of the credential definition
                from which to identify a schema
        """

        # scrape schema id or sequence number from cred def id
        tokens = credential_definition_id.split(":")
        if len(tokens) == 8:  # node protocol >= 1.4: cred def id has 5 or 8 tokens
            return ":".join(tokens[3:7])  # schema id spans 0-based positions 3-6

        # get txn by sequence number, retrieve schema identifier components
        seq_no = tokens[3]
        return (await self.get_schema(seq_no))["id"]

    async def get_key_for_did(self, did: str) -> Optional[str]:
        """Fetch the verkey for a ledger DID.

        Args:
            did: The DID to look up on the ledger or in the cache
        """
        nym = strip_did_prefix(did)
        public_info = await self.get_wallet_public_did()
        public_did = public_info.did if public_info else None

        # current public_did may be non-indy -> create nym request with empty public did
        if public_did is not None and not bool(IndyDID.PATTERN.match(public_did)):
            public_did = None

        try:
            nym_req = ledger.build_get_nym_request(public_did, nym)
        except VdrError as err:
            raise LedgerError("Exception when building get-nym request") from err

        response = await self._submit(nym_req, sign_did=public_info)
        data_json = response["data"]
        return json.loads(data_json)["verkey"] if data_json else None

    async def get_all_endpoints_for_did(self, did: str) -> dict:
        """Fetch all endpoints for a ledger DID.

        Args:
            did: The DID to look up on the ledger or in the cache
        """
        nym = strip_did_prefix(did)
        public_info = await self.get_wallet_public_did()
        public_did = public_info.did if public_info else None
        try:
            attrib_req = ledger.build_get_attrib_request(
                public_did, nym, "endpoint", None, None
            )
        except VdrError as err:
            raise LedgerError("Exception when building attribute request") from err

        response = await self._submit(attrib_req, sign_did=public_info)
        data_json = response["data"]

        if data_json:
            endpoints = json.loads(data_json).get("endpoint", None)
        else:
            endpoints = None

        return endpoints

    async def get_endpoint_for_did(
        self, did: str, endpoint_type: Optional[EndpointType] = None
    ) -> str:
        """Fetch the endpoint for a ledger DID.

        Args:
            did: The DID to look up on the ledger or in the cache
            endpoint_type: The type of the endpoint. If none given, returns all
        """

        if not endpoint_type:
            endpoint_type = EndpointType.ENDPOINT
        nym = strip_did_prefix(did)
        public_info = await self.get_wallet_public_did()
        public_did = public_info.did if public_info else None
        try:
            attrib_req = ledger.build_get_attrib_request(
                public_did, nym, "endpoint", None, None
            )
        except VdrError as err:
            raise LedgerError("Exception when building attribute request") from err

        response = await self._submit(attrib_req, sign_did=public_info)
        data_json = response["data"]
        if data_json:
            endpoint = json.loads(data_json).get("endpoint", None)
            address = endpoint.get(endpoint_type.indy, None) if endpoint else None
        else:
            address = None

        return address

    async def update_endpoint_for_did(
        self,
        did: str,
        endpoint: str,
        endpoint_type: Optional[EndpointType] = None,
        write_ledger: bool = True,
        endorser_did: Optional[str] = None,
        routing_keys: Optional[List[str]] = None,
    ) -> bool:
        """Check and update the endpoint on the ledger.

        Args:
            did: The ledger DID
            endpoint: The endpoint address
            endpoint_type: The type of the endpoint
            write_ledger: Whether to write the endpoint to the ledger
            endorser_did: DID of the endorser to use for the transaction
            routing_keys: List of routing keys
        """
        routing_keys = routing_keys or []  # Ensure list type if None was passed
        public_info = await self.get_wallet_public_did()
        if not public_info:
            raise BadLedgerRequestError(
                "Cannot update endpoint at ledger without a public DID"
            )

        if not endpoint_type:
            endpoint_type = EndpointType.ENDPOINT

        all_exist_endpoints = await self.get_all_endpoints_for_did(did)
        exist_endpoint_of_type = (
            all_exist_endpoints.get(endpoint_type.indy, None)
            if all_exist_endpoints
            else None
        )
        existing_routing_keys = (
            all_exist_endpoints.get("routingKeys") if all_exist_endpoints else None
        )
        existing_routing_keys = existing_routing_keys or []

        endpoint_changed = exist_endpoint_of_type != endpoint
        routing_keys_changed = set(existing_routing_keys) != set(routing_keys)

        if endpoint_changed or routing_keys_changed:
            LOGGER.info("Endpoint or routing keys have changed, updating endpoint")
            if self.read_only:
                raise LedgerError(
                    "Error cannot update endpoint when ledger is in read only mode"
                )

            nym = strip_did_prefix(did)

            attr_json = await self._construct_attr_json(
                endpoint, endpoint_type, all_exist_endpoints, routing_keys
            )

            try:
                attrib_req = ledger.build_attrib_request(nym, nym, None, attr_json, None)

                if endorser_did and not write_ledger:
                    attrib_req.set_endorser(endorser_did)
                    resp = await self._submit(
                        attrib_req,
                        True,
                        sign_did=public_info,
                        write_ledger=write_ledger,
                    )
                    if not write_ledger:
                        return {"signed_txn": resp}

            except VdrError as err:
                raise LedgerError("Exception when building attribute request") from err

            await self._submit(attrib_req, True, True)
            return True
        return False

    async def register_nym(
        self,
        did: str,
        verkey: str,
        alias: Optional[str] = None,
        role: Optional[str] = None,
        write_ledger: bool = True,
        endorser_did: Optional[str] = None,
    ) -> Tuple[bool, dict]:
        """Register a nym on the ledger.

        Args:
            did: DID to register on the ledger.
            verkey: The verification key of the keypair.
            alias: Human-friendly alias to assign to the DID.
            role: For permissioned ledgers, what role should the new DID have.
            write_ledger: Whether to write the nym to the ledger.
            endorser_did: DID of the endorser to use for the transaction.
        """
        if self.read_only:
            raise LedgerError(
                "Error cannot register nym when ledger is in read only mode"
            )

        public_info = await self.get_wallet_public_did()
        if not public_info:
            raise BadLedgerRequestError("Cannot register NYM without a public DID")

        try:
            nym_req = ledger.build_nym_request(public_info.did, did, verkey, alias, role)
        except VdrError as err:
            raise LedgerError("Exception when building nym request") from err

        if endorser_did and not write_ledger:
            nym_req.set_endorser(endorser_did)

        resp = await self._submit(
            nym_req, sign=True, sign_did=public_info, write_ledger=write_ledger
        )
        if not write_ledger:
            return True, {"signed_txn": resp}
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            try:
                did_info = await wallet.get_local_did(did)
            except WalletNotFoundError:
                pass  # not a local DID
            else:
                metadata = {**did_info.metadata, **DIDPosture.POSTED.metadata}
                await wallet.replace_local_did_metadata(did, metadata)
        return True, None

    async def get_nym_role(self, did: str) -> Role:
        """Return the role of the input public DID's NYM on the ledger.

        Args:
            did: DID to query for role on the ledger.
        """
        public_info = await self.get_wallet_public_did()
        public_did = public_info.did if public_info else None

        try:
            nym_req = ledger.build_get_nym_request(public_did, did)
        except VdrError as err:
            raise LedgerError("Exception when building get-nym request") from err

        response = await self._submit(nym_req)
        nym_data = json.loads(response["data"])
        if not nym_data:
            raise BadLedgerRequestError(f"DID {did} is not public")

        return Role.get(nym_data["role"])

    def nym_to_did(self, nym: str) -> str:
        """Format a nym with the ledger's DID prefix."""
        if nym:
            # remove any existing prefix
            nym = strip_did_prefix(nym)
            return f"did:sov:{nym}"

    async def build_and_return_get_nym_request(
        self, submitter_did: Optional[str], target_did: str
    ) -> str:
        """Build GET_NYM request and return request_json."""
        try:
            request_json = ledger.build_get_nym_request(submitter_did, target_did)
            return request_json
        except VdrError as err:
            raise LedgerError("Exception when building get-nym request") from err

    async def submit_get_nym_request(self, request_json: str) -> str:
        """Submit GET_NYM request to ledger and return response_json."""
        response_json = await self._submit(request_json)
        return response_json

    async def rotate_public_did_keypair(self, next_seed: Optional[str] = None) -> None:
        """Rotate keypair for public DID: create new key, submit to ledger, update wallet.

        Args:
            next_seed: seed for incoming ed25519 keypair (default random)
        """
        # generate new key
        async with self.profile.transaction() as txn:
            wallet = txn.inject(BaseWallet)
            public_info = await wallet.get_public_did()
            public_did = public_info.did
            verkey = await wallet.rotate_did_keypair_start(public_did, next_seed)
            del wallet
            await txn.commit()

        # fetch current nym info from ledger
        nym = strip_did_prefix(public_did)
        try:
            get_nym_req = ledger.build_get_nym_request(public_did, nym)
        except VdrError as err:
            raise LedgerError("Exception when building nym request") from err

        response = await self._submit(get_nym_req)
        data = json.loads(response["data"])
        if not data:
            raise BadLedgerRequestError(
                f"Ledger has no public DID for wallet {self.profile.name}"
            )
        seq_no = data["seqNo"]

        try:
            txn_req = ledger.build_get_txn_request(None, None, seq_no)
        except VdrError as err:
            raise LedgerError("Exception when building get-txn request") from err

        txn_resp = await self._submit(txn_req)
        txn_resp_data = txn_resp["data"]
        if not txn_resp_data:
            raise BadLedgerRequestError(
                f"Bad or missing ledger NYM transaction for DID {public_did}"
            )
        txn_data_data = txn_resp_data["txn"]["data"]
        role_token = Role.get(txn_data_data.get("role")).token()
        alias = txn_data_data.get("alias")
        # submit the updated nym record
        await self.register_nym(public_did, verkey, alias=alias, role=role_token)

        # update wallet
        async with self.profile.transaction() as txn:
            wallet = txn.inject(BaseWallet)
            await wallet.rotate_did_keypair_apply(public_did)
            del wallet
            await txn.commit()

    async def get_txn_author_agreement(self, reload: bool = False) -> dict:
        """Get the current transaction author agreement, fetching it if necessary."""
        if not self.pool.taa_cache or reload:
            self.pool.taa_cache = await self.fetch_txn_author_agreement()
        return self.pool.taa_cache

    async def fetch_txn_author_agreement(self) -> dict:
        """Fetch the current AML and TAA from the ledger."""
        public_info = await self.get_wallet_public_did()
        public_did = public_info.did if public_info else None

        get_aml_req = ledger.build_get_acceptance_mechanisms_request(
            public_did, None, None
        )
        response = await self._submit(get_aml_req, sign_did=public_info)
        aml_found = response["data"]

        get_taa_req = ledger.build_get_txn_author_agreement_request(public_did, None)
        response = await self._submit(get_taa_req, sign_did=public_info)
        taa_found = response["data"]
        taa_required = bool(taa_found and taa_found["text"])
        if taa_found:
            taa_found["digest"] = self.taa_digest(taa_found["version"], taa_found["text"])

        return {
            "aml_record": aml_found,
            "taa_record": taa_found,
            "taa_required": taa_required,
        }

    def taa_rough_timestamp(self) -> int:
        """Get a timestamp accurate to the day.

        Anything more accurate is a privacy concern.
        """
        return int(
            datetime.combine(date.today(), datetime.min.time(), timezone.utc).timestamp()
        )

    async def accept_txn_author_agreement(
        self, taa_record: dict, mechanism: str, accept_time: Optional[int] = None
    ):
        """Save a new record recording the acceptance of the TAA."""
        if not accept_time:
            accept_time = self.taa_rough_timestamp()
        acceptance = {
            "text": taa_record["text"],
            "version": taa_record["version"],
            "digest": taa_record["digest"],
            "mechanism": mechanism,
            "time": accept_time,
        }
        record = StorageRecord(
            TAA_ACCEPTED_RECORD_TYPE,
            json.dumps(acceptance),
            {"pool_name": self.pool_name},
        )
        async with self.profile.session() as session:
            storage = session.inject(BaseStorage)
            cache = self.profile.inject_or(BaseCache)
            await storage.add_record(record)
        if cache:
            cache_key = TAA_ACCEPTED_RECORD_TYPE + "::" + self.profile.name
            await cache.set(cache_key, acceptance, self.pool.cache_duration)

    async def get_latest_txn_author_acceptance(self) -> dict:
        """Look up the latest TAA acceptance."""
        cache_key = TAA_ACCEPTED_RECORD_TYPE + "::" + self.profile.name
        acceptance = self.pool.cache and await self.pool.cache.get(cache_key)
        if not acceptance:
            tag_filter = {"pool_name": self.pool_name}
            async with self.profile.session() as session:
                storage = session.inject(BaseStorage)
                cache = self.profile.inject_or(BaseCache)
                found = await storage.find_all_records(
                    TAA_ACCEPTED_RECORD_TYPE, tag_filter
                )
            if found:
                records = [json.loads(record.value) for record in found]
                records.sort(key=lambda v: v["time"], reverse=True)
                acceptance = records[0]
            else:
                acceptance = {}
            if cache:
                await cache.set(cache_key, acceptance, self.pool.cache_duration)
        return acceptance

    async def get_revoc_reg_def(self, revoc_reg_id: str) -> dict:
        """Get revocation registry definition by ID."""
        public_info = await self.get_wallet_public_did()
        try:
            fetch_req = ledger.build_get_revoc_reg_def_request(
                public_info and public_info.did, revoc_reg_id
            )
            response = await self._submit(fetch_req, sign_did=public_info)
        except VdrError as err:
            raise LedgerError(
                f"get_revoc_reg_def failed for revoc_reg_id='{revoc_reg_id}'"
            ) from err

        revoc_reg_def = response["data"]
        revoc_reg_def["ver"] = "1.0"
        revoc_reg_def["txnTime"] = response["txnTime"]

        if revoc_reg_def.get("id") != revoc_reg_id:
            raise LedgerError(
                "ID of revocation registry response does not match requested ID"
            )
        return revoc_reg_def

    async def get_revoc_reg_entry(
        self, revoc_reg_id: str, timestamp: int
    ) -> Tuple[dict, int]:
        """Get revocation registry entry by revocation registry ID and timestamp."""
        public_info = await self.get_wallet_public_did()
        try:
            fetch_req = ledger.build_get_revoc_reg_request(
                public_info and public_info.did, revoc_reg_id, timestamp
            )
            response = await self._submit(fetch_req, sign_did=public_info)
        except VdrError as err:
            raise LedgerError(
                f"get_revoc_reg_entry failed for revoc_reg_id='{revoc_reg_id}'"
            ) from err

        ledger_timestamp = response["data"]["txnTime"]
        reg_entry = {
            "ver": "1.0",
            "value": response["data"]["value"],
        }
        if response["data"]["revocRegDefId"] != revoc_reg_id:
            raise LedgerError(
                "ID of revocation registry response does not match requested ID"
            )
        return reg_entry, ledger_timestamp

    async def get_revoc_reg_delta(
        self, revoc_reg_id: str, timestamp_from=0, timestamp_to=None
    ) -> Tuple[dict, int]:
        """Look up a revocation registry delta by ID.

        :param revoc_reg_id revocation registry id
        :param timestamp_from from time. a total number of seconds from Unix Epoch
        :param timestamp_to to time. a total number of seconds from Unix Epoch

        :returns delta response, delta timestamp
        """
        if timestamp_to is None:
            timestamp_to = int(time())
        public_info = await self.get_wallet_public_did()
        try:
            fetch_req = ledger.build_get_revoc_reg_delta_request(
                public_info and public_info.did,
                revoc_reg_id,
                timestamp_from,
                timestamp_to,
            )
            response = await self._submit(fetch_req, sign_did=public_info)
        except VdrError as err:
            raise LedgerError(
                f"get_revoc_reg_delta failed for revoc_reg_id='{revoc_reg_id}'"
            ) from err

        response_value = response["data"]["value"]
        accum_to = response_value.get("accum_to")

        # If accum_to is not present, then the timestamp_to was before the registry
        # was created. In this case, we need to fetch the registry creation timestamp and
        # re-calculate the delta.
        if not accum_to:
            try:
                (_, timestamp) = await self.get_revoc_reg_entry(revoc_reg_id, int(time()))
                fetch_req = ledger.build_get_revoc_reg_delta_request(
                    public_info and public_info.did,
                    revoc_reg_id,
                    timestamp_from,
                    timestamp,
                )
                response = await self._submit(fetch_req, sign_did=public_info)
                response_value = response["data"]["value"]
            except VdrError as err:
                raise LedgerError(
                    f"get_revoc_reg_delta failed for revoc_reg_id='{revoc_reg_id}'"
                ) from err

        delta_value = {
            "accum": response_value["accum_to"]["value"]["accum"],
            "issued": response_value.get("issued", []),
            "revoked": response_value.get("revoked", []),
        }
        accum_from = response_value.get("accum_from")
        if accum_from:
            delta_value["prev_accum"] = accum_from["value"]["accum"]
        reg_delta = {"ver": "1.0", "value": delta_value}
        # question - why not response["to"] ?
        delta_timestamp = response_value["accum_to"]["txnTime"]
        if response["data"]["revocRegDefId"] != revoc_reg_id:
            raise LedgerError(
                "ID of revocation registry response does not match requested ID"
            )
        return reg_delta, delta_timestamp

    async def send_revoc_reg_def(
        self,
        revoc_reg_def: dict,
        issuer_did: Optional[str] = None,
        write_ledger: bool = True,
        endorser_did: Optional[str] = None,
    ) -> dict:
        """Publish a revocation registry definition to the ledger."""
        # NOTE - issuer DID could be extracted from the revoc_reg_def ID
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            if issuer_did:
                did_info = await wallet.get_local_did(issuer_did)
            else:
                did_info = await wallet.get_public_did()
            del wallet
        if not did_info:
            raise LedgerTransactionError(
                "No issuer DID found for revocation registry definition"
            )

        if self.profile.context.settings.get("wallet.type") == "askar-anoncreds":
            from acapy_agent.anoncreds.default.legacy_indy.registry import (
                LegacyIndyRegistry,
            )

            rev_reg_def_req = await self._create_revoc_reg_def_request(
                did_info,
                json.dumps(revoc_reg_def),
                write_ledger=write_ledger,
                endorser_did=endorser_did,
            )

            if endorser_did and not write_ledger:
                rev_reg_def_req.set_endorser(endorser_did)

            legacy_indy_registry = LegacyIndyRegistry()
            try:
                resp = await legacy_indy_registry.txn_submit(
                    self,
                    rev_reg_def_req,
                    sign=True,
                    sign_did=did_info,
                    write_ledger=write_ledger,
                )
            except LedgerError as err:
                raise LedgerError(
                    "Exception when sending revocation registry definition"
                ) from err

            if not write_ledger:
                return revoc_reg_def["id"], {"signed_txn": resp}

            try:
                # parse sequence number out of response
                seq_no = json.loads(resp)["result"]["txnMetadata"]["seqNo"]
                return seq_no
            except KeyError as err:
                raise LedgerError(
                    "Failed to parse sequence number from ledger response"
                ) from err
        else:
            try:
                request = ledger.build_revoc_reg_def_request(
                    did_info.did, json.dumps(revoc_reg_def)
                )
                if endorser_did and not write_ledger:
                    request.set_endorser(endorser_did)
            except VdrError as err:
                raise LedgerError(
                    "Exception when sending revocation registry definition"
                ) from err

            resp = await self._submit(
                request, True, sign_did=did_info, write_ledger=write_ledger
            )
            return {"result": resp}

    async def send_revoc_reg_entry(
        self,
        revoc_reg_id: str,
        revoc_def_type: str,
        revoc_reg_entry: dict,
        issuer_did: Optional[str] = None,
        write_ledger: bool = True,
        endorser_did: Optional[str] = None,
    ) -> dict:
        """Publish a revocation registry entry to the ledger."""
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            if issuer_did:
                did_info = await wallet.get_local_did(issuer_did)
            else:
                did_info = await wallet.get_public_did()
            del wallet
        if not did_info:
            raise LedgerTransactionError(
                "No issuer DID found for revocation registry entry"
            )

        if self.profile.context.settings.get("wallet.type") == "askar-anoncreds":
            from acapy_agent.anoncreds.default.legacy_indy.registry import (
                LegacyIndyRegistry,
            )

            revoc_reg_entry_req = await self._create_revoc_reg_entry_request(
                did_info,
                revoc_reg_id,
                revoc_def_type,
                json.dumps(revoc_reg_entry),
                write_ledger=write_ledger,
                endorser_did=endorser_did,
            )
            if endorser_did and not write_ledger:
                revoc_reg_entry_req.set_endorser(endorser_did)

            legacy_indy_registry = LegacyIndyRegistry()
            resp = await legacy_indy_registry.txn_submit(
                self,
                revoc_reg_entry_req,
                sign=True,
                sign_did=did_info,
                write_ledger=write_ledger,
            )

            if not write_ledger:
                return revoc_reg_id, {"signed_txn": resp}

            try:
                # parse sequence number out of response
                seq_no = json.loads(resp)["result"]["txnMetadata"]["seqNo"]
                return seq_no
            except KeyError as err:
                raise LedgerError(
                    "Failed to parse sequence number from ledger response"
                ) from err
        else:
            try:
                request = ledger.build_revoc_reg_entry_request(
                    did_info.did,
                    revoc_reg_id,
                    revoc_def_type,
                    json.dumps(revoc_reg_entry),
                )
                if endorser_did and not write_ledger:
                    request.set_endorser(endorser_did)
            except VdrError as err:
                raise LedgerError(
                    "Exception when sending revocation registry entry"
                ) from err
            resp = await self._submit(
                request, True, sign_did=did_info, write_ledger=write_ledger
            )
            return {"result": resp}

    async def get_wallet_public_did(self) -> DIDInfo:
        """Fetch the public DID from the wallet."""
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            return await wallet.get_public_did()

    async def txn_endorse(
        self,
        request_json: str,
        endorse_did: Optional[DIDInfo] = None,
    ) -> str:
        """Endorse (sign) the provided transaction."""
        try:
            request = ledger.build_custom_request(request_json)
        except VdrError as err:
            raise BadLedgerRequestError("Cannot endorse invalid request") from err

        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            sign_did = endorse_did if endorse_did else await wallet.get_public_did()
            if not sign_did:
                raise BadLedgerRequestError(
                    "Cannot endorse transaction without a public DID"
                )
            request.set_multi_signature(
                sign_did.did,
                await wallet.sign_message(request.signature_input, sign_did.verkey),
            )
            del wallet

        return request.body

    async def txn_submit(
        self,
        request_json: str,
        sign: bool,
        taa_accept: Optional[bool] = None,
        sign_did: DIDInfo = sentinel,
        write_ledger: bool = True,
    ) -> str:
        """Write the provided (signed and possibly endorsed) transaction to the ledger."""
        resp = await self._submit(
            request_json,
            sign=sign,
            taa_accept=taa_accept,
            sign_did=sign_did,
            write_ledger=write_ledger,
        )
        if write_ledger:
            # match the format returned by indy sdk
            resp = {"op": "REPLY", "result": resp}
        return json.dumps(resp)



"""Test universal resolver with http bindings."""

import re
from typing import Dict, Optional, Union

import pytest
import pytest_asyncio

from ....config.settings import Settings
from ....tests import mock
from ....utils.testing import create_test_profile
from ...base import DIDNotFound, ResolverError
from .. import universal as test_module
from ..universal import UniversalResolver


@pytest_asyncio.fixture
async def resolver():
    """Resolver fixture."""
    yield UniversalResolver(
        endpoint="https://example.com", supported_did_regex=re.compile("^did:sov:.*$")
    )


@pytest_asyncio.fixture
async def profile():
    """Profile fixture."""
    profile = await create_test_profile()
    yield profile


class MockResponse:
    """Mock http response."""

    def __init__(self, status: int, body: Union[str, Dict]):
        self.status = status
        self.body = body

    async def json(self):
        return self.body

    async def text(self):
        return self.body

    async def __aenter__(self):
        """For use as async context."""
        return self

    async def __aexit__(self, err_type, err_value, err_exc):
        """For use as async context."""


class MockClientSession:
    """Mock client session."""

    def __init__(self, response: Optional[MockResponse] = None):
        self.response = response

    def __call__(self, headers):
        return self

    async def __aenter__(self):
        """For use as async context."""
        return self

    async def __aexit__(self, err_type, err_value, err_exc):
        """For use as async context."""

    def get(self, endpoint):
        """Return response."""
        return self.response


@pytest.fixture
def mock_client_session():
    temp = test_module.aiohttp.ClientSession
    session = MockClientSession()
    test_module.aiohttp.ClientSession = session
    yield session
    test_module.aiohttp.ClientSession = temp


@pytest.mark.asyncio
async def test_resolve(profile, resolver, mock_client_session):
    mock_client_session.response = MockResponse(
        200,
        {
            "didDocument": {
                "id": "did:example:123",
                "@context": "https://www.w3.org/ns/did/v1",
            }
        },
    )
    doc = await resolver.resolve(profile, "did:sov:WRfXPg8dantKVubE3HX8pw")
    assert doc.get("id") == "did:example:123"


@pytest.mark.asyncio
async def test_resolve_not_found(profile, resolver, mock_client_session):
    mock_client_session.response = MockResponse(404, "Not found")
    with pytest.raises(DIDNotFound):
        await resolver.resolve(profile, "did:sov:WRfXPg8dantKVubE3HX8pw")


@pytest.mark.asyncio
async def test_resolve_unexpected_status(profile, resolver, mock_client_session):
    mock_client_session.response = MockResponse(500, "Server failed to complete request")
    with pytest.raises(ResolverError):
        await resolver.resolve(profile, "did:sov:WRfXPg8dantKVubE3HX8pw")


@pytest.mark.asyncio
async def test_fetch_resolver_props(mock_client_session: MockClientSession):
    mock_client_session.response = MockResponse(200, {"test": "json"})
    assert await UniversalResolver()._fetch_resolver_props() == {"test": "json"}
    mock_client_session.response = MockResponse(404, "Not found")
    with pytest.raises(ResolverError):
        await UniversalResolver()._fetch_resolver_props()


@pytest.mark.asyncio
async def test_get_supported_did_regex():
    # Old response format
    props = {"example": {"http": {"pattern": "match a test string"}}}
    with mock.patch.object(
        UniversalResolver,
        "_fetch_resolver_props",
        mock.CoroutineMock(return_value=props),
    ):
        pattern = await UniversalResolver()._get_supported_did_regex()
        assert pattern.fullmatch("match a test string")

    # Example response from dev universal resolver 1.0
    props = {
        "^(did:sov:(?:(?:\\w[-\\w]*(?::\\w[-\\w]*)*):)?(?:[1-9A-HJ-NP-Za-km-z]{21,22}))$": {
            "libIndyPath": "",
            "openParallel": "false",
            "poolVersions": "_;2;test;2;builder;2;danube;2;idunion;2;idunion:test;2;indicio;2;indicio:test;2;indicio:demo;2;nxd;2;findy:test;2;bcovrin;2;bcovrin:test;2;bcovrin:dev;2;candy;2;candy:test;2;candy:dev;2",
            "submitterDidSeeds": "_;_;test;_;builder;_;danube;_;idunion;_;idunion:test;_;indicio;_;indicio:test;_;indicio:demo;_;nxd;_;findy:test;_;bcovrin;_;bcovrin:test;_;bcovrin:dev;_;candy;_;candy:test;_;candy:dev;_",
            "http": {
                "resolveUri": "http://driver-did-sov:8080/1.0/identifiers/",
                "propertiesUri": "http://driver-did-sov:8080/1.0/properties",
            },
            "walletNames": "_;w1;test;w2;builder;w3;danube;w4;idunion;w5;idunion:test;w6;indicio;w7;indicio:test;w8;indicio:demo;w9;nxd;w11;findy:test;w12;bcovrin;w13;bcovrin:test;w14;bcovrin:dev;w15;candy;w16;candy:test;w17;candy:dev;w18",
            "poolConfigs": "_;./sovrin/_.txn;test;./sovrin/test.txn;builder;./sovrin/builder.txn;danube;./sovrin/danube.txn;idunion;./sovrin/idunion.txn;idunion:test;./sovrin/idunion-test.txn;indicio;./sovrin/indicio.txn;indicio:test;./sovrin/indicio-test.txn;indicio:demo;./sovrin/indicio-demo.txn;nxd;./sovrin/nxd.txn;bcovrin:test;./sovrin/bcovrin-test.txn;candy;./sovrin/candy.txn;candy:test;./sovrin/candy-test.txn;candy:dev;./sovrin/candy-dev.txn",
        }
    }
    with mock.patch.object(
        UniversalResolver,
        "_fetch_resolver_props",
        mock.CoroutineMock(return_value=props),
    ):
        pattern = await UniversalResolver()._get_supported_did_regex()
        assert pattern.match("did:sov:WRfXPg8dantKVubE3HX8pw")


def test_compile_supported_did_regex():
    patterns = ["one", "two", "three"]
    compiled = test_module._compile_supported_did_regex(patterns)
    assert compiled.match("one")
    assert compiled.match("two")
    assert compiled.match("three")


@pytest.mark.asyncio
async def test_setup_endpoint_regex_set(resolver: UniversalResolver):
    settings = Settings(
        {
            "resolver.universal": "http://example.com",
            "resolver.universal.supported": "test",
        }
    )
    context = mock.MagicMock()
    context.settings = settings
    with mock.patch.object(
        test_module,
        "_compile_supported_did_regex",
        mock.MagicMock(return_value="pattern"),
    ):
        await resolver.setup(context)

    assert resolver._endpoint == "http://example.com"
    assert resolver._supported_did_regex == "pattern"


@pytest.mark.asyncio
async def test_setup_endpoint_set(resolver: UniversalResolver):
    settings = Settings(
        {
            "resolver.universal": "http://example.com",
        }
    )
    context = mock.MagicMock()
    context.settings = settings
    with mock.patch.object(
        UniversalResolver,
        "_get_supported_did_regex",
        mock.CoroutineMock(return_value="pattern"),
    ):
        await resolver.setup(context)

    assert resolver._endpoint == "http://example.com"
    assert resolver._supported_did_regex == "pattern"


@pytest.mark.asyncio
async def test_setup_endpoint_default(resolver: UniversalResolver):
    settings = Settings(
        {
            "resolver.universal": "DEFAULT",
        }
    )
    context = mock.MagicMock()
    context.settings = settings
    with mock.patch.object(
        UniversalResolver,
        "_get_supported_did_regex",
        mock.CoroutineMock(return_value="pattern"),
    ):
        await resolver.setup(context)

    assert resolver._endpoint == test_module.DEFAULT_ENDPOINT
    assert resolver._supported_did_regex == "pattern"


@pytest.mark.asyncio
async def test_setup_endpoint_unset(resolver: UniversalResolver):
    settings = Settings()
    context = mock.MagicMock()
    context.settings = settings
    with mock.patch.object(
        UniversalResolver,
        "_get_supported_did_regex",
        mock.CoroutineMock(return_value="pattern"),
    ):
        await resolver.setup(context)

    assert resolver._endpoint == test_module.DEFAULT_ENDPOINT
    assert resolver._supported_did_regex == "pattern"


@pytest.mark.asyncio
async def test_supported_did_regex_not_setup():
    resolver = UniversalResolver()
    with pytest.raises(ResolverError):
        resolver.supported_did_regex



from unittest import IsolatedAsyncioTestCase

from ......core.oob_processor import OobMessageProcessor
from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......tests import mock
from ......transport.inbound.receipt import MessageReceipt
from ......utils.testing import create_test_profile
from ...messages.cred_issue import V20CredIssue
from .. import cred_issue_handler as test_module


class TestCredentialIssueHandler(IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.request_context = RequestContext.test_context(await create_test_profile())
        self.request_context.message_receipt = MessageReceipt()
        self.request_context.settings["debug.auto_store_credential"] = False
        self.request_context.connection_record = mock.MagicMock()
        self.mock_oob_processor = mock.MagicMock(OobMessageProcessor, autospec=True)
        self.mock_oob_processor.find_oob_record_for_inbound_message = mock.CoroutineMock(
            return_value=mock.MagicMock()
        )
        self.request_context.injector.bind_instance(
            OobMessageProcessor, self.mock_oob_processor
        )

    async def test_called(self):
        with mock.patch.object(
            test_module, "V20CredManager", autospec=True
        ) as mock_cred_mgr:
            mock_cred_mgr.return_value.receive_credential = mock.CoroutineMock()
            self.request_context.message = V20CredIssue()
            self.request_context.connection_ready = True
            handler_inst = test_module.V20CredIssueHandler()
            responder = MockResponder()
            await handler_inst.handle(self.request_context, responder)

        mock_cred_mgr.assert_called_once_with(self.request_context.profile)
        mock_cred_mgr.return_value.receive_credential.assert_called_once_with(
            self.request_context.message,
            self.request_context.connection_record.connection_id,
        )
        self.mock_oob_processor.find_oob_record_for_inbound_message.assert_called_once_with(
            self.request_context
        )
        assert not responder.messages

    async def test_called_auto_store(self):
        with mock.patch.object(
            test_module, "V20CredManager", autospec=True
        ) as mock_cred_mgr:
            mock_cred_mgr.return_value = mock.MagicMock(
                receive_credential=mock.CoroutineMock(),
                store_credential=mock.CoroutineMock(),
                send_cred_ack=mock.CoroutineMock(return_value="cred_ack_message"),
            )
            self.request_context.message = V20CredIssue()
            self.request_context.connection_ready = True
            handler_inst = test_module.V20CredIssueHandler()
            responder = MockResponder()
            self.request_context.settings["debug.auto_store_credential"] = True
            await handler_inst.handle(self.request_context, responder)

        mock_cred_mgr.assert_called_once_with(self.request_context.profile)
        mock_cred_mgr.return_value.receive_credential.assert_called_once_with(
            self.request_context.message,
            self.request_context.connection_record.connection_id,
        )
        self.mock_oob_processor.find_oob_record_for_inbound_message.assert_called_once_with(
            self.request_context
        )
        assert mock_cred_mgr.return_value.send_cred_ack.call_count == 1

    async def test_called_auto_store_x_indy(self):
        with mock.patch.object(
            test_module, "V20CredManager", autospec=True
        ) as mock_cred_mgr:
            mock_cred_mgr.return_value = mock.MagicMock(
                receive_credential=mock.CoroutineMock(
                    return_value=mock.MagicMock(save_error_state=mock.CoroutineMock())
                ),
                store_credential=mock.CoroutineMock(
                    side_effect=[
                        test_module.IndyHolderError,
                        test_module.StorageError(),
                    ]
                ),
                send_cred_ack=mock.CoroutineMock(),
            )

            self.request_context.message = V20CredIssue()
            self.request_context.connection_ready = True
            handler_inst = test_module.V20CredIssueHandler()
            responder = MockResponder()

            await handler_inst.handle(self.request_context, responder)  # holder error
            await handler_inst.handle(self.request_context, responder)  # storage error

    async def test_called_auto_store_x_anoncreds(self):
        with mock.patch.object(
            test_module, "V20CredManager", autospec=True
        ) as mock_cred_mgr:
            mock_cred_mgr.return_value = mock.MagicMock(
                receive_credential=mock.CoroutineMock(
                    return_value=mock.MagicMock(save_error_state=mock.CoroutineMock())
                ),
                store_credential=mock.CoroutineMock(
                    side_effect=[
                        test_module.AnonCredsHolderError,
                        test_module.StorageError(),
                    ]
                ),
                send_cred_ack=mock.CoroutineMock(),
            )

            self.request_context.message = V20CredIssue()
            self.request_context.connection_ready = True
            handler_inst = test_module.V20CredIssueHandler()
            responder = MockResponder()

            await handler_inst.handle(self.request_context, responder)  # holder error
            await handler_inst.handle(self.request_context, responder)  # storage error

    async def test_called_not_ready(self):
        with mock.patch.object(
            test_module, "V20CredManager", autospec=True
        ) as mock_cred_mgr:
            mock_cred_mgr.return_value.receive_credential = mock.CoroutineMock()
            self.request_context.message = V20CredIssue()
            self.request_context.connection_ready = False
            handler_inst = test_module.V20CredIssueHandler()
            responder = MockResponder()
            with self.assertRaises(test_module.HandlerException) as err:
                await handler_inst.handle(self.request_context, responder)
            assert err.exception.message == "Connection used for credential not ready"

        assert not responder.messages

    async def test_called_no_connection_no_oob(self):
        self.request_context.message = V20CredIssue()
        handler_inst = test_module.V20CredIssueHandler()
        responder = MockResponder()
        self.request_context.connection_ready = True
        self.request_context.connection_record = None

        self.mock_oob_processor = mock.MagicMock(OobMessageProcessor, autospec=True)
        self.mock_oob_processor.find_oob_record_for_inbound_message = mock.CoroutineMock(
            return_value=None
        )
        self.request_context.injector.bind_instance(
            OobMessageProcessor, self.mock_oob_processor
        )
        with self.assertRaises(test_module.HandlerException) as err:
            await handler_inst.handle(self.request_context, responder)
        assert (
            err.exception.message
            == "No connection or associated connectionless exchange found for credential"
        )

        assert not responder.messages



TEST_LD_DOCUMENT_BAD_SIGNED_P256_2019 = {
    "@context": ["http://schema.org/", "https://w3id.org/security/bbs/v1"],
    "@type": "Person",
    "firstName": "Jane",
    "lastName": "Does",
    "jobTitle": "Professor",
    "telephone": "(425) 123-4567",
    "email": "jane.doe@example.com",
    "proof": {
        "type": "EcdsaSecp256r1Signature2019",
        "verificationMethod": "did:key:zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq#zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq",
        "created": "2025-01-15T00:55:37+00:00",
        "proofPurpose": "assertionMethod",
        "proofValue": "z42hBVFNcjBLGpdMcDpA9YfPagJ9Ky5Fv8PQf7KoRt3S8tYJq61shtnsSszdLxQRpospia8CrM76rKE2MRRQimQKb",
    },
}



"""Settings routes."""

import logging

from aiohttp import web
from aiohttp_apispec import docs, request_schema, response_schema
from marshmallow import fields

from ..admin.request_context import AdminRequestContext
from ..core.error import BaseError
from ..core.profile import Profile
from ..messaging.models.openapi import OpenAPISchema
from ..multitenant.admin.routes import (
    ACAPY_LIFECYCLE_CONFIG_FLAG_ARGS_MAP,
    get_extra_settings_dict_per_tenant,
)
from ..multitenant.base import BaseMultitenantManager

LOGGER = logging.getLogger(__name__)


class UpdateProfileSettingsSchema(OpenAPISchema):
    """Schema to update profile settings."""

    extra_settings = fields.Dict(
        required=False,
        metadata={
            "description": "Agent config key-value pairs",
            "example": {
                "log-level": "INFO",
                "ACAPY_INVITE_PUBLIC": True,
                "public-invites": False,
            },
        },
    )


class ProfileSettingsSchema(OpenAPISchema):
    """Profile settings response schema."""

    settings = fields.Dict(
        metadata={
            "description": "Profile settings dict",
            "example": {
                "log.level": "INFO",
                "debug.invite_public": True,
                "public_invites": False,
            },
        }
    )


def _get_filtered_settings_dict(wallet_settings: dict):
    """Get filtered settings dict to display."""
    filter_param_list = list(ACAPY_LIFECYCLE_CONFIG_FLAG_ARGS_MAP.values())
    filter_param_list.append("endorser.author")
    filter_param_list.append("endorser.endorser")
    filter_param_list.append("wallet.type")
    settings_dict = {}
    for param in filter_param_list:
        if param in wallet_settings:
            settings_dict[param] = wallet_settings.get(param)
    return settings_dict


def _get_multitenant_settings_dict(
    profile_settings: dict,
    wallet_settings: dict,
):
    """Get filtered settings dict when multitenant manager is present."""
    all_settings = {**profile_settings, **wallet_settings}
    settings_dict = _get_filtered_settings_dict(all_settings)
    return settings_dict


def _get_settings_dict(
    profile: Profile,
):
    """Get filtered settings dict when multitenant manager is not present."""
    settings_dict = _get_filtered_settings_dict((profile.settings).to_dict())
    return settings_dict


@docs(
    tags=["settings"],
    summary="Update configurable settings associated with the profile.",
)
@request_schema(UpdateProfileSettingsSchema())
@response_schema(ProfileSettingsSchema(), 200, description="")
async def update_profile_settings(request: web.BaseRequest):
    """Request handler for updating setting associated with profile.

    Args:
        request: aiohttp request object
    """
    context: AdminRequestContext = request["context"]
    root_profile = context.root_profile or context.profile
    try:
        body = await request.json()
        extra_settings = get_extra_settings_dict_per_tenant(
            body.get("extra_settings") or {}
        )
        async with root_profile.session() as session:
            multitenant_mgr = session.inject_or(BaseMultitenantManager)
            if multitenant_mgr:
                wallet_id = context.metadata.get("wallet_id")
                wallet_key = context.metadata.get("wallet_key")
                wallet_record = await multitenant_mgr.update_wallet(
                    wallet_id, extra_settings
                )
                wallet_record, profile = await multitenant_mgr.get_wallet_and_profile(
                    root_profile.context, wallet_id, wallet_key
                )
                settings_dict = _get_multitenant_settings_dict(
                    profile_settings=profile.settings.to_dict(),
                    wallet_settings=wallet_record.settings,
                )
            else:
                root_profile.context.update_settings(extra_settings)
                settings_dict = _get_settings_dict(profile=root_profile)
    except BaseError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err
    return web.json_response(settings_dict)


@docs(
    tags=["settings"],
    summary="Get the configurable settings associated with the profile.",
)
@response_schema(ProfileSettingsSchema(), 200, description="")
async def get_profile_settings(request: web.BaseRequest):
    """Request handler for getting setting associated with profile.

    Args:
        request: aiohttp request object
    """
    context: AdminRequestContext = request["context"]
    root_profile = context.root_profile or context.profile
    try:
        async with root_profile.session() as session:
            multitenant_mgr = session.inject_or(BaseMultitenantManager)
            if multitenant_mgr:
                wallet_id = context.metadata.get("wallet_id")
                wallet_key = context.metadata.get("wallet_key")
                wallet_record, profile = await multitenant_mgr.get_wallet_and_profile(
                    root_profile.context, wallet_id, wallet_key
                )
                settings_dict = _get_multitenant_settings_dict(
                    profile_settings=profile.settings.to_dict(),
                    wallet_settings=wallet_record.settings,
                )
            else:
                settings_dict = _get_settings_dict(profile=root_profile)
    except BaseError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err
    return web.json_response(settings_dict)


async def register(app: web.Application):
    """Register routes."""

    app.add_routes(
        [
            web.put("/settings", update_profile_settings),
            web.get("/settings", get_profile_settings, allow_head=False),
        ]
    )


def post_process_routes(app: web.Application):
    """Amend swagger API."""

    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {
            "name": "settings",
            "description": "Agent settings interface.",
        }
    )



import pytest
import pytest_asyncio

from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......transport.inbound.receipt import MessageReceipt
from ......utils.testing import create_test_profile
from ......wallet.did_method import DIDMethods
from .....out_of_band.v1_0.messages.invitation import InvitationMessage
from ...handlers.invitation_handler import InvitationHandler
from ...messages.problem_report import DIDXProblemReport, ProblemReportReason


@pytest_asyncio.fixture
async def request_context():
    ctx = RequestContext.test_context(await create_test_profile())
    ctx.injector.bind_instance(DIDMethods, DIDMethods())
    ctx.message_receipt = MessageReceipt()
    yield ctx


class TestDIDXInvitationHandler:
    @pytest.mark.asyncio
    async def test_problem_report(self, request_context):
        request_context.message = InvitationMessage()
        handler = InvitationHandler()
        responder = MockResponder()
        await handler.handle(request_context, responder)
        messages = responder.messages
        assert len(messages) == 1
        result, target = messages[0]
        assert (
            isinstance(result, DIDXProblemReport)
            and result.description
            and (
                result.description["code"]
                == ProblemReportReason.INVITATION_NOT_ACCEPTED.value
            )
        )
        assert not target



"""Revoke message."""

from typing import Optional

from marshmallow import fields

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from .....messaging.valid import UUID4_EXAMPLE
from ..message_types import PROTOCOL_PACKAGE, REVOKE

HANDLER_CLASS = f"{PROTOCOL_PACKAGE}.handlers.revoke_handler.RevokeHandler"


class Revoke(AgentMessage):
    """Class representing revoke message."""

    class Meta:
        """Revoke Meta."""

        handler_class = HANDLER_CLASS
        message_type = REVOKE
        schema_class = "RevokeSchema"

    def __init__(self, *, thread_id: str, comment: Optional[str] = None, **kwargs):
        """Initialize revoke message."""
        super().__init__(**kwargs)
        # TODO support please ack
        self.thread_id = thread_id
        self.comment = comment


class RevokeSchema(AgentMessageSchema):
    """Schema of Revoke message."""

    class Meta:
        """RevokeSchema Meta."""

        model_class = Revoke

    # TODO support please ack
    thread_id = fields.Str(
        required=True,
        metadata={
            "description": (
                "Thread ID of credential exchange resulting in this issued credential"
            ),
            "example": UUID4_EXAMPLE,
        },
    )
    comment = fields.Str(
        required=False,
        metadata={
            "description": "Human readable information about revocation notification"
        },
    )



"""An object for containing information on an individual route."""

from typing import Optional

from marshmallow import EXCLUDE, ValidationError, fields, validates_schema

from .....core.profile import ProfileSession
from .....messaging.models.base_record import BaseRecord, BaseRecordSchema


class RouteRecord(BaseRecord):
    """Class representing stored route information."""

    class Meta:
        """RouteRecord metadata."""

        schema_class = "RouteRecordSchema"

    RECORD_TYPE = "forward_route"
    RECORD_ID_NAME = "record_id"
    ROLE_CLIENT = "client"
    ROLE_SERVER = "server"
    TAG_NAMES = {"connection_id", "role", "recipient_key", "wallet_id"}

    def __init__(
        self,
        *,
        record_id: Optional[str] = None,
        role: Optional[str] = None,
        connection_id: Optional[str] = None,
        wallet_id: Optional[str] = None,
        recipient_key: Optional[str] = None,
        **kwargs,
    ):
        """Initialize route record.

        Args:
            record_id (str): record_id optionally specify record id manually
            role (str): role of agent, client or server
            connection_id (str): connection_id associated with record
            wallet_id: The id of the wallet for the route. Used for multitenant relay
            recipient_key (str): recipient_key associated with record
            kwargs: additional args for BaseRecord
        """
        super().__init__(record_id, None, **kwargs)
        self.role = role or self.ROLE_SERVER
        self.connection_id = connection_id
        self.wallet_id = wallet_id
        self.recipient_key = recipient_key

    def __eq__(self, other: "RouteRecord"):
        """Equality check."""
        if not isinstance(other, RouteRecord):
            return False
        return (
            self.record_id == other.record_id
            and self.record_tags == other.record_tags
            and self.record_value == other.record_value
        )

    @property
    def record_id(self) -> str:
        """Get record ID."""
        return self._id

    @classmethod
    async def retrieve_by_recipient_key(
        cls, session: ProfileSession, recipient_key: str
    ) -> "RouteRecord":
        """Retrieve a route record by recipient key.

        Args:
            session (ProfileSession): session
            recipient_key (str): key to look up

        Returns:
            RouteRecord: retrieved route record

        """
        tag_filter = {"recipient_key": recipient_key}
        return await cls.retrieve_by_tag_filter(session, tag_filter)

    @classmethod
    async def retrieve_by_connection_id(
        cls, session: ProfileSession, connection_id: str
    ) -> "RouteRecord":
        """Retrieve a route record by connection ID.

        Args:
            session (ProfileSession): session
            connection_id (str): ID to look up

        Returns:
            RouteRecord: retrieved route record

        """
        tag_filter = {"connection_id": connection_id}
        return await cls.retrieve_by_tag_filter(session, tag_filter)

    @property
    def record_value(self) -> dict:
        """Accessor for JSON record value."""
        return {
            prop: getattr(self, prop)
            for prop in (
                "connection_id",
                "wallet_id",
                "recipient_key",
            )
        }


class RouteRecordSchema(BaseRecordSchema):
    """RouteRecord schema."""

    class Meta:
        """RouteRecordSchema metadata."""

        model_class = RouteRecord
        unknown = EXCLUDE

    record_id = fields.Str()
    role = fields.Str(required=False)
    connection_id = fields.Str()
    wallet_id = fields.Str()
    recipient_key = fields.Str(required=True)

    @validates_schema
    def validate_fields(self, data, **kwargs):
        """Validate schema fields.

        Args:
            data: The data to validate
            kwargs: Additional keyword arguments

        Raises:
            ValidationError: If any of the fields do not validate

        """

        if not (data.get("connection_id") or data.get("wallet_id")):
            raise ValidationError(
                "Either connection_id or wallet_id must be set for route"
            )



"""Record used to represent a service block of an out of band invitation."""

from typing import Optional, Sequence

from marshmallow import EXCLUDE, fields, post_dump

from .....messaging.models.base import BaseModel, BaseModelSchema
from .....messaging.valid import (
    DID_KEY_OR_REF_EXAMPLE,
    DID_KEY_OR_REF_VALIDATE,
    INDY_DID_EXAMPLE,
    INDY_DID_VALIDATE,
)


class Service(BaseModel):
    """Record used to represent a service block of an out of band invitation."""

    class Meta:
        """Service metadata."""

        schema_class = "ServiceSchema"

    def __init__(
        self,
        *,
        _id: Optional[str] = None,
        _type: Optional[str] = None,
        did: Optional[str] = None,
        recipient_keys: Optional[Sequence[str]] = None,
        routing_keys: Optional[Sequence[str]] = None,
        service_endpoint: Optional[str] = None,
    ):
        """Initialize a Service instance.

        Args:
            _id: An identifier for this service block
            _type: A type for this service block
            did: A did for the connection
            recipient_keys: A list of recipient keys in W3C did:key format
            routing_keys: A list of routing keys in W3C did:key format
            service_endpoint: An endpoint for the connection
        """
        self._id = _id
        self._type = _type
        self.did = did
        self.recipient_keys = list(recipient_keys) if recipient_keys else []
        self.routing_keys = list(routing_keys) if routing_keys else []
        self.service_endpoint = service_endpoint


class ServiceSchema(BaseModelSchema):
    """Service schema."""

    class Meta:
        """ServiceSchema metadata."""

        model_class = Service
        unknown = EXCLUDE

    _id = fields.Str(
        required=True, data_key="id", metadata={"description": "Service identifier"}
    )
    _type = fields.Str(
        required=True, data_key="type", metadata={"description": "Service type"}
    )
    did = fields.Str(
        required=False,
        validate=INDY_DID_VALIDATE,
        metadata={"description": "Service DID", "example": INDY_DID_EXAMPLE},
    )

    recipient_keys = fields.List(
        fields.Str(
            validate=DID_KEY_OR_REF_VALIDATE,
            metadata={
                "description": "Recipient public key",
                "example": DID_KEY_OR_REF_EXAMPLE,
            },
        ),
        data_key="recipientKeys",
        required=False,
        metadata={"description": "List of recipient keys"},
    )

    routing_keys = fields.List(
        fields.Str(
            validate=DID_KEY_OR_REF_VALIDATE,
            metadata={"description": "Routing key", "example": DID_KEY_OR_REF_EXAMPLE},
        ),
        data_key="routingKeys",
        required=False,
        metadata={"description": "List of routing keys"},
    )

    service_endpoint = fields.Str(
        data_key="serviceEndpoint",
        required=False,
        metadata={
            "description": "Service endpoint at which to reach this agent",
            "example": "http://192.168.56.101:8020",
        },
    )

    @post_dump
    def post_dump(self, data, **kwargs):
        """Post dump hook."""

        if "routingKeys" in data and not data["routingKeys"]:
            del data["routingKeys"]

        return data



"""BBS+ crypto."""

from typing import List, Optional, Tuple

from ..core.error import BaseError
from ..utils.dependencies import (
    assert_ursa_bbs_signatures_installed,
    is_ursa_bbs_signatures_module_installed,
)
from ..wallet.util import random_seed

if is_ursa_bbs_signatures_module_installed():
    from ursa_bbs_signatures import BbsException as NativeBbsException
    from ursa_bbs_signatures import BlsKeyPair, SignRequest, VerifyRequest
    from ursa_bbs_signatures import sign as bbs_sign
    from ursa_bbs_signatures import verify as bbs_verify
    from ursa_bbs_signatures._ffi.FfiException import FfiException


class BbsException(BaseError):
    """Base BBS exception."""


def sign_messages_bls12381g2(messages: List[bytes], secret: bytes):
    """Sign messages using a bls12381g2 private signing key.

    Args:
        messages (List[bytes]): The messages to sign
        secret (bytes): The private signing key

    Returns:
        bytes: The signature

    """
    assert_ursa_bbs_signatures_installed()

    messages = [message.decode("utf-8") for message in messages]
    try:
        key_pair = BlsKeyPair.from_secret_key(secret)
        sign_request = SignRequest(key_pair=key_pair, messages=messages)
    except (
        FfiException,
        NativeBbsException,
    ) as error:  # would be nice to be able to distinct between false and error
        raise BbsException("Unable to sign messages") from error

    return bbs_sign(sign_request)


def verify_signed_messages_bls12381g2(
    messages: List[bytes], signature: bytes, public_key: bytes
) -> bool:
    """Verify an ed25519 signed message according to a public verification key.

    Args:
        messages (List[bytes]): The signed messages to verify.
        signature (bytes): The signature to verify.
        public_key (bytes): The public key to use in verification.

    Returns:
        bool: True if the signature is verified, else False.

    Raises:
        BbsException: If unable to verify the BBS+ signature.

    """
    assert_ursa_bbs_signatures_installed()

    key_pair = BlsKeyPair(public_key=public_key)
    messages = [message.decode("utf-8") for message in messages]

    verify_request = VerifyRequest(
        key_pair=key_pair, signature=signature, messages=messages
    )

    try:
        return bbs_verify(verify_request)
    except (
        FfiException,
        NativeBbsException,
    ) as error:
        raise BbsException("Unable to verify BBS+ signature") from error


def create_bls12381g2_keypair(seed: Optional[bytes] = None) -> Tuple[bytes, bytes]:
    """Create a public and private bls12381g2 keypair from a seed value.

    Args:
        seed: Seed for keypair

    Returns:
        A tuple of (public key, secret key)

    """
    assert_ursa_bbs_signatures_installed()

    if not seed:
        seed = random_seed()

    try:
        key_pair = BlsKeyPair.generate_g2(seed)
        return key_pair.public_key, key_pair.secret_key
    except Exception as error:
        raise BbsException("Unable to create keypair") from error



"""Represents a trust ping message."""

from typing import Optional

from marshmallow import EXCLUDE, fields

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from ..message_types import PING, PROTOCOL_PACKAGE

HANDLER_CLASS = f"{PROTOCOL_PACKAGE}.handlers.ping_handler.PingHandler"


class Ping(AgentMessage):
    """Class representing a trustping message."""

    class Meta:
        """Ping metadata."""

        handler_class = HANDLER_CLASS
        message_type = PING
        schema_class = "PingSchema"

    def __init__(
        self, *, response_requested: bool = True, comment: Optional[str] = None, **kwargs
    ):
        """Initialize a Ping message instance.

        Args:
            response_requested: A flag indicating that a response is requested
                (defaults to True for the recipient if not included)
            comment: An optional comment string
            kwargs: Keyword arguments for message construction

        """
        super().__init__(**kwargs)
        self.comment = comment
        self.response_requested = response_requested


class PingSchema(AgentMessageSchema):
    """Schema for Ping class."""

    class Meta:
        """PingSchema metadata."""

        model_class = Ping
        unknown = EXCLUDE

    response_requested = fields.Bool(
        dump_default=True,
        required=False,
        metadata={
            "description": "Whether response is requested (default True)",
            "example": True,
        },
    )
    comment = fields.Str(
        required=False,
        allow_none=True,
        metadata={"description": "Optional comment to include", "example": "Hello"},
    )



from unittest import IsolatedAsyncioTestCase, mock

from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......transport.inbound.receipt import MessageReceipt
from ......utils.testing import create_test_profile
from ...messages.ack import V10Ack
from .. import ack_handler as test_module


class TestNotificationAckHandler(IsolatedAsyncioTestCase):
    async def test_called(self):
        request_context = RequestContext.test_context(await create_test_profile())
        request_context.message_receipt = MessageReceipt()
        request_context.connection_record = mock.MagicMock()
        request_context.connection_ready = True

        request_context.message = V10Ack(status="OK")
        handler = test_module.V10AckHandler()
        responder = MockResponder()
        await handler.handle(request_context, responder)



"""Module to contain logic to generate the banner for ACA-py."""

import logging
import textwrap
from contextlib import contextmanager
from enum import Enum, auto
from typing import Optional, TextIO

logger = logging.getLogger(__name__)


@contextmanager
def Banner(border: str, length: int, file: Optional[TextIO] = None):
    """Context manager to generate a banner for ACA-py."""
    banner = _Banner(border, length, file)
    banner.add_border()
    yield banner
    banner.add_border()

    # Join all lines with newlines and log them
    banner_text = "\n".join(banner.lines)
    banner_text = f"\n{banner_text.strip()}\n"  # Start/end with a newline
    if file:
        print(banner_text, file=file)
    else:
        logger.info(banner_text)


class _Banner:
    """Management class to generate a banner for ACA-py."""

    class align(Enum):
        """Alignment options for banner elements."""

        LEFT = auto()
        CENTER = auto()
        RIGHT = auto()

    def __init__(self, border: str, length: int, file: Optional[TextIO] = None):
        """Initialize the banner object.

        The ``border`` is a single character to be used, and ``length``
        is the desired length of the whole banner, inclusive.
        """
        self.border = border
        self.length = length
        self.file = file
        self.lines = []

    def _add_line(self, text: str):
        """Add a line to the banner."""
        self.lines.append(text)

    def _lr_pad(self, content: str):
        """Pad string content with defined border character.

        Args:
            content: String content to pad
        """
        return f"{self.border}{self.border} {content} {self.border}{self.border}"

    def _format_line(self, text: str, alignment: align = align.LEFT):
        """Format a line with the specified alignment."""
        lines = textwrap.wrap(text, width=self.length)
        formatted_lines = []

        for line in lines:
            if len(line) < self.length:
                if alignment == self.align.LEFT:
                    # Left alignment
                    formatted_line = f"{line:<{self.length}}"
                elif alignment == self.align.CENTER:
                    # Center alignment
                    total_padding = self.length - len(line)
                    left_padding = total_padding // 2
                    right_padding = total_padding - left_padding
                    formatted_line = f"{' ' * left_padding}{line}{' ' * right_padding}"
                elif alignment == self.align.RIGHT:
                    # Right alignment
                    formatted_line = f"{line:>{self.length}}"
                else:
                    raise ValueError(f"Invalid alignment: {alignment}")
            else:
                formatted_line = line

            formatted_lines.append(self._lr_pad(formatted_line))

        return formatted_lines

    def add_border(self):
        """Add a full line using the border character."""
        self._add_line(self.border * (self.length + 6))

    def title(self, title, spacing_after: int = 2):
        """Add the main title element."""
        self.lines.extend(self._format_line(title, self.align.CENTER))
        for _ in range(spacing_after):
            self.spacer()

    def spacer(self):
        """Add an empty line with the border character only."""
        self._add_line(self._lr_pad(" " * self.length))

    def hr(self, char: str = "-"):
        """Add a line with a horizontal rule."""
        self._add_line(self._lr_pad(char * self.length))

    def subtitle(self, title: str, spacing_after: int = 1):
        """Add a subtitle for a section."""
        title += ":"
        self.lines.extend(self._format_line(title, self.align.LEFT))
        for _ in range(spacing_after):
            self.spacer()

    def list(self, items, spacing_after: int = 1):
        """Add a list of items, prepending a dash to each item."""
        for item in items:
            self.lines.extend(self._format_line(f"  - {item}", self.align.LEFT))

        for _ in range(spacing_after):
            self.spacer()

    def version(self, version):
        """Add the current ``version``."""
        version = f"ver: {version}"
        self.lines.extend(self._format_line(version, self.align.RIGHT))

    def print(self, text: str):
        """Add a line of text."""
        self.lines.extend(self._format_line(text, self.align.LEFT))

    def left(self, text: str):
        """Add a line of text left aligned.

        Same as `print` method.
        """
        self.lines.extend(self._format_line(text, self.align.LEFT))

    def centered(self, text: str):
        """Add a line of text centered."""
        self.lines.extend(self._format_line(text, self.align.CENTER))

    def right(self, text: str):
        """Add a line of text right aligned."""
        self.lines.extend(self._format_line(text, self.align.RIGHT))



"""Credential artifacts."""

from typing import Mapping, Optional

from marshmallow import EXCLUDE, ValidationError, fields

from ...messaging.models.base import BaseModel, BaseModelSchema
from ...messaging.valid import (
    INDY_CRED_DEF_ID_EXAMPLE,
    INDY_CRED_DEF_ID_VALIDATE,
    INDY_REV_REG_ID_EXAMPLE,
    INDY_REV_REG_ID_VALIDATE,
    INDY_SCHEMA_ID_EXAMPLE,
    INDY_SCHEMA_ID_VALIDATE,
    NUM_STR_ANY_EXAMPLE,
    NUM_STR_ANY_VALIDATE,
)


class IndyAttrValue(BaseModel):
    """Indy attribute value."""

    class Meta:
        """Indy attribute value."""

        schema_class = "IndyAttrValueSchema"

    def __init__(
        self, raw: Optional[str] = None, encoded: Optional[str] = None, **kwargs
    ):
        """Initialize indy (credential) attribute value."""
        super().__init__(**kwargs)
        self.raw = raw
        self.encoded = encoded


class IndyAttrValueSchema(BaseModelSchema):
    """Indy attribute value schema."""

    class Meta:
        """Indy attribute value schema metadata."""

        model_class = IndyAttrValue
        unknown = EXCLUDE

    raw = fields.Str(required=True, metadata={"description": "Attribute raw value"})
    encoded = fields.Str(
        required=True,
        validate=NUM_STR_ANY_VALIDATE,
        metadata={
            "description": "Attribute encoded value",
            "example": NUM_STR_ANY_EXAMPLE,
        },
    )


class DictWithIndyAttrValueSchema(fields.Dict):
    """Dict with indy attribute value schema."""

    def __init__(self, **kwargs):
        """Initialize the custom schema for a dictionary with IndyAttrValue."""
        super().__init__(
            keys=fields.Str(metadata={"description": "Attribute name"}),
            values=fields.Nested(IndyAttrValueSchema()),
            **kwargs,
        )

    def _deserialize(self, value, attr, data, **kwargs):
        """Deserialize dict with indy attribute value."""
        if not isinstance(value, dict):
            raise ValidationError("Value must be a dict.")

        errors = {}
        indy_attr_value_schema = IndyAttrValueSchema()

        for k, v in value.items():
            if isinstance(v, dict):
                validation_errors = indy_attr_value_schema.validate(v)
                if validation_errors:
                    errors[k] = validation_errors

        if errors:
            raise ValidationError(errors)

        return value


class IndyCredential(BaseModel):
    """Indy credential."""

    class Meta:
        """Indy credential metadata."""

        schema_class = "IndyCredentialSchema"

    def __init__(
        self,
        schema_id: Optional[str] = None,
        cred_def_id: Optional[str] = None,
        rev_reg_id: Optional[str] = None,
        values: Mapping[str, IndyAttrValue] = None,
        signature: Optional[Mapping] = None,
        signature_correctness_proof: Optional[Mapping] = None,
        rev_reg: Optional[Mapping] = None,
        witness: Optional[Mapping] = None,
    ):
        """Initialize indy credential."""
        self.schema_id = schema_id
        self.cred_def_id = cred_def_id
        self.rev_reg_id = rev_reg_id
        self.values = values
        self.signature = signature
        self.signature_correctness_proof = signature_correctness_proof
        self.rev_reg = rev_reg
        self.witness = witness


class IndyCredentialSchema(BaseModelSchema):
    """Indy credential schema."""

    class Meta:
        """Indy credential schemametadata."""

        model_class = IndyCredential
        unknown = EXCLUDE

    schema_id = fields.Str(
        required=True,
        validate=INDY_SCHEMA_ID_VALIDATE,
        metadata={
            "description": "Schema identifier",
            "example": INDY_SCHEMA_ID_EXAMPLE,
        },
    )
    cred_def_id = fields.Str(
        required=True,
        validate=INDY_CRED_DEF_ID_VALIDATE,
        metadata={
            "description": "Credential definition identifier",
            "example": INDY_CRED_DEF_ID_EXAMPLE,
        },
    )
    rev_reg_id = fields.Str(
        allow_none=True,
        validate=INDY_REV_REG_ID_VALIDATE,
        metadata={
            "description": "Revocation registry identifier",
            "example": INDY_REV_REG_ID_EXAMPLE,
        },
    )
    values = DictWithIndyAttrValueSchema(
        required=True,
        metadata={"description": "Credential attributes"},
    )
    signature = fields.Dict(
        required=True, metadata={"description": "Credential signature"}
    )
    signature_correctness_proof = fields.Dict(
        required=True,
        metadata={"description": "Credential signature correctness proof"},
    )
    rev_reg = fields.Dict(
        allow_none=True, metadata={"description": "Revocation registry state"}
    )
    witness = fields.Dict(
        allow_none=True, metadata={"description": "Witness for revocation proof"}
    )



"""Credential request message."""

from typing import Optional, Sequence

from marshmallow import EXCLUDE, ValidationError, fields, validates_schema

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from .....messaging.decorators.attach_decorator import (
    AttachDecorator,
    AttachDecoratorSchema,
)
from ..message_types import CRED_20_REQUEST, PROTOCOL_PACKAGE
from .cred_format import V20CredFormat, V20CredFormatSchema

HANDLER_CLASS = f"{PROTOCOL_PACKAGE}.handlers.cred_request_handler.V20CredRequestHandler"


class V20CredRequest(AgentMessage):
    """Credential request."""

    class Meta:
        """V20CredRequest metadata."""

        handler_class = HANDLER_CLASS
        schema_class = "V20CredRequestSchema"
        message_type = CRED_20_REQUEST

    def __init__(
        self,
        _id: Optional[str] = None,
        *,
        comment: Optional[str] = None,
        formats: Sequence[V20CredFormat] = None,
        requests_attach: Sequence[AttachDecorator] = None,
        **kwargs,
    ):
        """Initialize credential request object.

        Args:
            _id (str): The identifier for the credential request.
            comment (str, optional): An optional comment for the credential request.
            formats (Sequence[V20CredFormat], optional): A list of acceptable attachment
                formats.
            requests_attach (Sequence[AttachDecorator], optional): A list of request
                attachments.
            kwargs: Additional key-value arguments to map into message class properties.

        """
        super().__init__(_id=_id, **kwargs)
        self.comment = comment
        self.formats = list(formats) if formats else []
        self.requests_attach = list(requests_attach) if requests_attach else []

    def attachment(self, fmt: V20CredFormat.Format = None) -> dict:
        """Return attached credential request.

        Args:
            fmt: format of attachment in list to decode and return

        """
        target_format = (
            fmt
            if fmt
            else next(
                filter(
                    lambda ff: ff,
                    [V20CredFormat.Format.get(f.format) for f in self.formats],
                ),
                None,
            )
        )
        return (
            target_format.get_attachment_data(self.formats, self.requests_attach)
            if target_format
            else None
        )


class V20CredRequestSchema(AgentMessageSchema):
    """Credential request schema."""

    class Meta:
        """Credential request schema metadata."""

        model_class = V20CredRequest
        unknown = EXCLUDE

    comment = fields.Str(
        required=False,
        allow_none=True,
        metadata={"description": "Human-readable comment"},
    )
    formats = fields.Nested(
        V20CredFormatSchema,
        many=True,
        required=True,
        metadata={"description": "Acceptable attachment formats"},
    )
    requests_attach = fields.Nested(
        AttachDecoratorSchema,
        required=True,
        many=True,
        data_key="requests~attach",
        metadata={"description": "Request attachments"},
    )

    @validates_schema
    def validate_fields(self, data, **kwargs):
        """Validate attachments per format."""

        def get_attach_by_id(attach_id):
            """Return attachment with input identifier."""
            for atch in attachments:
                if atch.ident == attach_id:
                    return atch
            raise ValidationError(f"No attachment for attach_id {attach_id} in formats")

        formats = data.get("formats") or []
        attachments = data.get("requests_attach") or []
        if len(formats) != len(attachments):
            raise ValidationError("Formats/attachments length mismatch")

        for fmt in formats:
            atch = get_attach_by_id(fmt.attach_id)
            cred_format = V20CredFormat.Format.get(fmt.format)

            if cred_format:
                cred_format.validate_fields(CRED_20_REQUEST, atch.content)



"""Class to provide some common utilities.

For Connection, DIDExchange and OutOfBand Manager.
"""

import json
import logging
import warnings
from typing import Dict, List, Optional, Sequence, Text, Tuple, Union

import pydid
from base58 import b58decode
from did_peer_2 import KeySpec, generate
from did_peer_4 import encode, long_to_short
from did_peer_4.input_doc import KeySpec as KeySpec_DP4
from did_peer_4.input_doc import input_doc_from_keys_and_services
from pydid import BaseDIDDocument as ResolvedDocument
from pydid import DIDCommService, VerificationMethod
from pydid.verification_method import (
    Ed25519VerificationKey2018,
    Ed25519VerificationKey2020,
    JsonWebKey2020,
    Multikey,
)

from ..cache.base import BaseCache
from ..config.base import InjectionError
from ..core.error import BaseError
from ..core.profile import Profile
from ..did.did_key import DIDKey
from ..multitenant.base import BaseMultitenantManager
from ..protocols.coordinate_mediation.v1_0.models.mediation_record import MediationRecord
from ..protocols.coordinate_mediation.v1_0.route_manager import RouteManager
from ..protocols.didexchange.v1_0.message_types import ARIES_PROTOCOL as CONN_PROTO
from ..protocols.discovery.v2_0.manager import V20DiscoveryMgr
from ..protocols.out_of_band.v1_0.messages.invitation import InvitationMessage
from ..resolver.base import ResolverError
from ..resolver.did_resolver import DIDResolver
from ..storage.base import BaseStorage
from ..storage.error import StorageDuplicateError, StorageNotFoundError
from ..storage.record import StorageRecord
from ..transport.inbound.receipt import MessageReceipt
from ..utils.multiformats import multibase, multicodec
from ..wallet.base import BaseWallet
from ..wallet.crypto import create_keypair, seed_to_did
from ..wallet.did_info import INVITATION_REUSE_KEY, DIDInfo, KeyInfo
from ..wallet.did_method import PEER2, PEER4, SOV, DIDMethod
from ..wallet.error import WalletNotFoundError
from ..wallet.key_type import ED25519, X25519
from ..wallet.util import b64_to_bytes, bytes_to_b58
from .models.conn_record import ConnRecord
from .models.connection_target import ConnectionTarget
from .models.diddoc import DIDDoc, PublicKey, PublicKeyType, Service


class BaseConnectionManagerError(BaseError):
    """BaseConnectionManager error."""


class BaseConnectionManager:
    """Class to provide utilities regarding connection_targets."""

    RECORD_TYPE_DID_DOC = "did_doc"
    RECORD_TYPE_DID_KEY = "did_key"

    def __init__(self, profile: Profile):
        """Initialize a BaseConnectionManager.

        Args:
            profile (Profile): The profile object associated with this connection manager.

        """
        self._profile = profile
        self._route_manager = profile.inject(RouteManager)
        self._logger = logging.getLogger(__name__)

    @staticmethod
    def _key_info_to_multikey(key_info: KeyInfo) -> str:
        """Convert a KeyInfo to a multikey."""
        if key_info.key_type == ED25519:
            return multibase.encode(
                multicodec.wrap("ed25519-pub", b58decode(key_info.verkey)), "base58btc"
            )
        elif key_info.key_type == X25519:
            return multibase.encode(
                multicodec.wrap("x25519-pub", b58decode(key_info.verkey)), "base58btc"
            )
        else:
            raise BaseConnectionManagerError(
                "Unsupported key type. Could not convert to multikey."
            )

    def long_did_peer_to_short(self, long_did: str) -> str:
        """Convert did:peer:4 long format to short format and return."""

        short_did_peer = long_to_short(long_did)
        return short_did_peer

    async def long_did_peer_4_to_short(self, long_dp4: str) -> str:
        """Convert did:peer:4 long format to short format and store in wallet."""

        async with self._profile.session() as session:
            wallet = session.inject(BaseWallet)
            long_dp4_info = await wallet.get_local_did(long_dp4)

        short_did_peer_4 = long_to_short(long_dp4)
        did_info = DIDInfo(
            did=short_did_peer_4,
            method=PEER4,
            verkey=long_dp4_info.verkey,
            metadata={},
            key_type=ED25519,
        )
        async with self._profile.session() as session:
            wallet = session.inject(BaseWallet)
            await wallet.store_did(did_info)
        return did_info.did

    async def create_did_peer_4(
        self,
        svc_endpoints: Optional[Sequence[str]] = None,
        mediation_records: Optional[List[MediationRecord]] = None,
        metadata: Optional[Dict] = None,
    ) -> DIDInfo:
        """Create a did:peer:4 DID for a connection.

        Args:
            svc_endpoints (Optional[Sequence[str]]): Custom endpoints for the
                DID Document.
            mediation_records (Optional[List[MediationRecord]]): The records for
                mediation that contain routing keys and service endpoint.
            metadata (Optional[Dict]): Additional metadata for the DID.

        Returns:
            DIDInfo: The new `DIDInfo` instance representing the created DID.
        """
        routing_keys: List[str] = []
        if mediation_records:
            for mediation_record in mediation_records:
                (
                    mediator_routing_keys,
                    endpoint,
                ) = await self._route_manager.routing_info(
                    self._profile, mediation_record
                )
                routing_keys = [*routing_keys, *(mediator_routing_keys or [])]
                if endpoint:
                    svc_endpoints = [endpoint]

        services = []
        for index, endpoint in enumerate(svc_endpoints or []):
            services.append(
                {
                    "id": f"#didcomm-{index}",
                    "type": "did-communication",
                    "recipientKeys": ["#key-0"],
                    "routingKeys": routing_keys,
                    "serviceEndpoint": endpoint,
                    "priority": index,
                }
            )

        async with self._profile.session() as session:
            wallet = session.inject(BaseWallet)
            key = await wallet.create_key(ED25519)
            key_spec = KeySpec_DP4(
                multikey=self._key_info_to_multikey(key),
                relationships=["authentication", "keyAgreement"],
            )
            input_doc = input_doc_from_keys_and_services(
                keys=[key_spec], services=services
            )
            did = encode(input_doc)

            did_metadata = metadata if metadata else {}
            did_info = DIDInfo(
                did=did,
                method=PEER4,
                verkey=key.verkey,
                metadata=did_metadata,
                key_type=ED25519,
            )
            await wallet.store_did(did_info)

        return did_info

    async def create_did_peer_2(
        self,
        svc_endpoints: Optional[Sequence[str]] = None,
        mediation_records: Optional[List[MediationRecord]] = None,
        metadata: Optional[Dict] = None,
    ) -> DIDInfo:
        """Create a did:peer:2 DID for a connection.

        Args:
            svc_endpoints (Optional[Sequence[str]]): Custom endpoints for the
                DID Document.
            mediation_records (Optional[List[MediationRecord]]): The records for
                mediation that contain routing keys and service endpoint.
            metadata (Optional[Dict]): Additional metadata for the DID.

        Returns:
            DIDInfo: The new `DIDInfo` instance representing the created DID.
        """
        routing_keys: List[str] = []
        if mediation_records:
            for mediation_record in mediation_records:
                (
                    mediator_routing_keys,
                    endpoint,
                ) = await self._route_manager.routing_info(
                    self._profile, mediation_record
                )
                routing_keys = [*routing_keys, *(mediator_routing_keys or [])]
                if endpoint:
                    svc_endpoints = [endpoint]

        services = []
        for index, endpoint in enumerate(svc_endpoints or []):
            services.append(
                {
                    "id": f"#didcomm-{index}",
                    "type": "did-communication",
                    "priority": index,
                    "recipientKeys": ["#key-1"],
                    "routingKeys": routing_keys,
                    "serviceEndpoint": endpoint,
                }
            )
            if self._profile.settings.get("experiment.didcomm_v2"):
                services.append(
                    {
                        "id": f"#service-{index}",
                        "type": "DIDCommMessaging",
                        "serviceEndpoint": {
                            "uri": endpoint,
                            "accept": ["didcomm/v2"],
                            "routingKeys": routing_keys,
                        },
                    }
                )

        async with self._profile.session() as session:
            wallet = session.inject(BaseWallet)
            key = await wallet.create_key(ED25519)
            xk = await wallet.create_key(X25519)

            did = generate(
                [
                    KeySpec.verification(self._key_info_to_multikey(key)),
                    KeySpec.key_agreement(self._key_info_to_multikey(xk)),
                ],
                services,
            )

            did_metadata = metadata if metadata else {}
            did_info = DIDInfo(
                did=did,
                method=PEER2,
                verkey=key.verkey,
                metadata=did_metadata,
                key_type=ED25519,
            )
            await wallet.store_did(did_info)
            await wallet.assign_kid_to_key(key.verkey, f"{did}#key-1")
            await wallet.assign_kid_to_key(xk.verkey, f"{did}#key-2")

        return did_info

    async def fetch_invitation_reuse_did(
        self,
        did_method: DIDMethod,
    ) -> Optional[DIDInfo]:
        """Fetch a DID from the wallet to use across multiple invitations.

        Args:
            did_method: The DID method used (e.g. PEER2 or PEER4)

        Returns:
            The `DIDInfo` instance, or "None" if no DID is found
        """
        did_info = None
        async with self._profile.session() as session:
            wallet = session.inject(BaseWallet)
            # TODO Iterating through all DIDs is problematic
            did_list = await wallet.get_local_dids()
            for did in did_list:
                if did.method == did_method and INVITATION_REUSE_KEY in did.metadata:
                    return did
        return did_info

    async def create_did_document(
        self,
        did_info: DIDInfo,
        svc_endpoints: Optional[Sequence[str]] = None,
        mediation_records: Optional[List[MediationRecord]] = None,
    ) -> DIDDoc:
        """Create our DID doc for a given DID.

        This method is deprecated and will be removed.

        Args:
            did_info (DIDInfo): The DID information (DID and verkey) used in the
                connection.
            svc_endpoints (Optional[Sequence[str]]): Custom endpoints for the
                DID Document.
            mediation_records (Optional[List[MediationRecord]]): The records for
                mediation that contain routing keys and service endpoints.

        Returns:
            DIDDoc: The prepared `DIDDoc` instance.

        """
        warnings.warn("create_did_document is deprecated and will be removed soon")
        did_doc = DIDDoc(did=did_info.did)
        did_controller = did_info.did
        did_key = did_info.verkey
        pk = PublicKey(
            did_info.did,
            "1",
            did_key,
            PublicKeyType.ED25519_SIG_2018,
            did_controller,
            True,
        )
        did_doc.set(pk)

        routing_keys: List[str] = []
        if mediation_records:
            for mediation_record in mediation_records:
                (
                    mediator_routing_keys,
                    endpoint,
                ) = await self._route_manager.routing_info(
                    self._profile, mediation_record
                )
                routing_keys = [*routing_keys, *(mediator_routing_keys or [])]
                if endpoint:
                    svc_endpoints = [endpoint]

        for endpoint_index, svc_endpoint in enumerate(svc_endpoints or []):
            endpoint_ident = "indy" if endpoint_index == 0 else f"indy{endpoint_index}"
            service = Service(
                did_info.did,
                endpoint_ident,
                "IndyAgent",
                [pk],
                routing_keys,
                svc_endpoint,
            )
            did_doc.set(service)

        return did_doc

    async def store_did_document(self, value: Union[DIDDoc, dict]):
        """Store a DID document.

        Args:
            value: The `DIDDoc` instance to persist
        """
        if isinstance(value, DIDDoc):
            did = value.did
            doc = value.to_json()
        else:
            did = value["id"]
            doc = json.dumps(value)

        # Special case: we used to store did:sov dids as unqualified.
        # For backwards compatibility, we'll strip off the prefix.
        if did.startswith("did:sov:"):
            did = did[8:]

        self._logger.debug("Storing DID document for %s: %s", did, doc)

        try:
            stored_doc, record = await self.fetch_did_document(did)
        except StorageNotFoundError:
            record = StorageRecord(self.RECORD_TYPE_DID_DOC, doc, {"did": did})
            async with self._profile.session() as session:
                storage: BaseStorage = session.inject(BaseStorage)
                await storage.add_record(record)
        else:
            async with self._profile.session() as session:
                storage: BaseStorage = session.inject(BaseStorage)
                await storage.update_record(record, doc, {"did": did})

        await self.remove_keys_for_did(did)
        await self.record_keys_for_resolvable_did(did)

    async def add_key_for_did(self, did: str, key: str):
        """Store a verkey for lookup against a DID.

        Args:
            did: The DID to associate with this key
            key: The verkey to be added
        """
        record = StorageRecord(self.RECORD_TYPE_DID_KEY, key, {"did": did, "key": key})
        async with self._profile.session() as session:
            storage: BaseStorage = session.inject(BaseStorage)
            try:
                await storage.find_record(self.RECORD_TYPE_DID_KEY, {"key": key})
            except StorageNotFoundError:
                await storage.add_record(record)
            except StorageDuplicateError:
                self._logger.warning(
                    "Key already associated with DID: %s; this is likely caused by "
                    "routing keys being erroneously stored in the past",
                    key,
                )

    async def find_did_for_key(self, key: str) -> str:
        """Find the DID previously associated with a key.

        Args:
            key: The verkey to look up
        """
        async with self._profile.session() as session:
            storage: BaseStorage = session.inject(BaseStorage)
            record = await storage.find_record(self.RECORD_TYPE_DID_KEY, {"key": key})
        ret_did = record.tags["did"]
        return ret_did

    async def remove_keys_for_did(self, did: str):
        """Remove all keys associated with a DID.

        Args:
            did: The DID for which to remove keys
        """
        async with self._profile.session() as session:
            storage: BaseStorage = session.inject(BaseStorage)
            await storage.delete_all_records(self.RECORD_TYPE_DID_KEY, {"did": did})

    async def resolve_didcomm_services(
        self, did: str, service_accept: Optional[Sequence[Text]] = None
    ) -> Tuple[ResolvedDocument, List[DIDCommService]]:
        """Resolve a DIDComm services for a given DID."""
        if not did.startswith("did:"):
            # DID is bare indy "nym"
            # prefix with did:sov: for backwards compatibility
            did = f"did:sov:{did}"

        resolver = self._profile.inject(DIDResolver)
        try:
            doc_dict: dict = await resolver.resolve(self._profile, did, service_accept)
            doc: ResolvedDocument = pydid.deserialize_document(doc_dict, strict=True)
        except (ResolverError, ValueError) as error:
            raise BaseConnectionManagerError("Failed to resolve DID services") from error

        if not doc.service:
            raise BaseConnectionManagerError(
                "Cannot connect via DID that has no associated services"
            )

        didcomm_services = sorted(
            [service for service in doc.service if isinstance(service, DIDCommService)],
            key=lambda service: service.priority,
        )

        return doc, didcomm_services

    async def verification_methods_for_service(
        self, doc: ResolvedDocument, service: DIDCommService
    ) -> Tuple[List[VerificationMethod], List[VerificationMethod]]:
        """Dereference recipient and routing keys.

        Returns verification methods for a DIDComm service to enable extracting
        key material.
        """
        self._logger.debug(
            "Getting recipient and routing keys for service %s", service.id
        )
        resolver = self._profile.inject(DIDResolver)
        recipient_keys: List[VerificationMethod] = [
            await resolver.dereference_verification_method(
                self._profile, url, document=doc
            )
            for url in service.recipient_keys
        ]
        routing_keys: List[VerificationMethod] = [
            await resolver.dereference_verification_method(
                self._profile, url, document=doc
            )
            for url in service.routing_keys
        ]
        return recipient_keys, routing_keys

    async def resolve_invitation(
        self, did: str, service_accept: Optional[Sequence[Text]] = None
    ) -> Tuple[str, List[str], List[str]]:
        """Resolve invitation with the DID Resolver.

        Args:
            did (str): Document ID to resolve.
            service_accept (Optional[Sequence[Text]]): List of accepted service types.

        Returns:
            Tuple[str, List[str], List[str]]: A tuple containing the endpoint,
                recipient keys, and routing keys.

        Raises:
            BaseConnectionManagerError: If the public DID has no associated
                DIDComm services.
        """
        self._logger.debug("Resolving invitation for DID %s", did)
        doc, didcomm_services = await self.resolve_didcomm_services(did, service_accept)
        if not didcomm_services:
            self._logger.warning("No DIDComm services found for DID %s", did)
            raise BaseConnectionManagerError(
                "Cannot connect via public DID that has no associated DIDComm services"
            )
        first_didcomm_service, *_ = didcomm_services
        self._logger.debug(
            "DIDComm service (id %s) found for DID %s", first_didcomm_service.id, did
        )

        endpoint = str(first_didcomm_service.service_endpoint)
        recipient_keys, routing_keys = await self.verification_methods_for_service(
            doc, first_didcomm_service
        )
        self._logger.debug(
            "DID %s has recipient keys %s and routing keys %s",
            did,
            recipient_keys,
            routing_keys,
        )
        return (
            endpoint,
            [self._extract_key_material_in_base58_format(key) for key in recipient_keys],
            [self._extract_key_material_in_base58_format(key) for key in routing_keys],
        )

    async def record_keys_for_resolvable_did(self, did: str):
        """Record the keys for a public DID.

        This is required to correlate sender verkeys back to a connection.
        """
        doc, didcomm_services = await self.resolve_didcomm_services(did)
        for service in didcomm_services:
            recips, _ = await self.verification_methods_for_service(doc, service)
            for recip in recips:
                await self.add_key_for_did(
                    did, self._extract_key_material_in_base58_format(recip)
                )

    async def resolve_connection_targets(
        self,
        did: str,
        sender_verkey: Optional[str] = None,
        their_label: Optional[str] = None,
    ) -> List[ConnectionTarget]:
        """Resolve connection targets for a DID."""
        self._logger.debug("Resolving connection targets for DID %s", did)
        doc, didcomm_services = await self.resolve_didcomm_services(did)
        self._logger.debug("Resolved DID document: %s", doc)
        self._logger.debug("Resolved DIDComm services: %s", didcomm_services)
        targets = []
        for service in didcomm_services:
            try:
                recips, routing = await self.verification_methods_for_service(
                    doc, service
                )
                endpoint = str(service.service_endpoint)
                targets.append(
                    ConnectionTarget(
                        did=doc.id,
                        endpoint=endpoint,
                        label=their_label,
                        recipient_keys=[
                            self._extract_key_material_in_base58_format(key)
                            for key in recips
                        ],
                        routing_keys=[
                            self._extract_key_material_in_base58_format(key)
                            for key in routing
                        ],
                        sender_key=sender_verkey,
                    )
                )
            except ResolverError:
                self._logger.exception(
                    "Failed to resolve service details while determining "
                    "connection targets; skipping service"
                )
                continue

        return targets

    @staticmethod
    def _extract_key_material_in_base58_format(method: VerificationMethod) -> str:
        if isinstance(method, Ed25519VerificationKey2018):
            return method.material
        elif isinstance(method, Ed25519VerificationKey2020):
            raw_data = multibase.decode(method.material)
            if len(raw_data) == 32:  # No multicodec prefix
                return bytes_to_b58(raw_data)
            else:
                codec, key = multicodec.unwrap(raw_data)
                if codec == multicodec.multicodec("ed25519-pub"):
                    return bytes_to_b58(key)
                else:
                    raise BaseConnectionManagerError(
                        f"Key type {type(method).__name__} "
                        f"with multicodec value {codec} is not supported"
                    )

        elif isinstance(method, JsonWebKey2020):
            if method.public_key_jwk.get("kty") == "OKP":
                return bytes_to_b58(b64_to_bytes(method.public_key_jwk.get("x"), True))
            else:
                raise BaseConnectionManagerError(
                    f"Key type {type(method).__name__} "
                    f"with kty {method.public_key_jwk.get('kty')} is not supported"
                )
        elif isinstance(method, Multikey):
            codec, key = multicodec.unwrap(multibase.decode(method.material))
            if codec != multicodec.multicodec("ed25519-pub"):
                raise BaseConnectionManagerError(
                    "Expected ed25519 multicodec, got: %s", codec
                )
            return bytes_to_b58(key)
        else:
            raise BaseConnectionManagerError(
                f"Key type {type(method).__name__} is not supported"
            )

    async def _fetch_connection_targets_for_invitation(
        self,
        connection: ConnRecord,
        invitation: InvitationMessage,
        sender_verkey: str,
    ) -> Sequence[ConnectionTarget]:
        """Get a list of connection targets for an invitation.

        This method extracts target information for either a connection or out-of-band
            (OOB) invitation.

        Args:
            connection (ConnRecord): The connection record associated with the invitation.
            invitation (InvitationMessage): The connection
                or OOB invitation retrieved from the connection record.
            sender_verkey (str): The sender's verification key.

        Returns:
            Sequence[ConnectionTarget]: A list of `ConnectionTarget` objects
                representing the connection targets for the invitation.
        """
        assert invitation.services, "Schema requires services in invitation"
        oob_service_item = invitation.services[0]
        if isinstance(oob_service_item, str):
            (
                endpoint,
                recipient_keys,
                routing_keys,
            ) = await self.resolve_invitation(oob_service_item)

        else:
            endpoint = oob_service_item.service_endpoint
            recipient_keys = [
                DIDKey.from_did(k).public_key_b58 for k in oob_service_item.recipient_keys
            ]
            routing_keys = [
                DIDKey.from_did(k).public_key_b58 for k in oob_service_item.routing_keys
            ]

        return [
            ConnectionTarget(
                did=connection.their_did,
                endpoint=endpoint,
                label=invitation.label if invitation else None,
                recipient_keys=recipient_keys,
                routing_keys=routing_keys,
                sender_key=sender_verkey,
            )
        ]

    async def _fetch_targets_for_connection_in_progress(
        self, connection: ConnRecord, sender_verkey: str
    ) -> Sequence[ConnectionTarget]:
        """Get a list of connection targets from an incomplete `ConnRecord`.

        This covers retrieving targets for connections that are still in the
        process of bootstrapping. This includes connections that are in states
        invitation-received or request-received.

        Args:
            connection: The connection record (with associated `DIDDoc`)
                used to generate the connection target
            sender_verkey: The verkey we are using
        Returns:
            A list of `ConnectionTarget` objects
        """
        if (
            connection.invitation_msg_id
            or connection.invitation_key
            or not connection.their_did
        ):  # invitation received or sending request to invitation
            async with self._profile.session() as session:
                invitation = await connection.retrieve_invitation(session)
            targets = await self._fetch_connection_targets_for_invitation(
                connection,
                invitation,
                sender_verkey,
            )
        else:  # sending implicit request
            # request is implicit; did isn't set if we've received an
            # invitation, only the invitation key
            (
                endpoint,
                recipient_keys,
                routing_keys,
            ) = await self.resolve_invitation(connection.their_did)
            targets = [
                ConnectionTarget(
                    did=connection.their_did,
                    endpoint=endpoint,
                    label=None,
                    recipient_keys=recipient_keys,
                    routing_keys=routing_keys,
                    sender_key=sender_verkey,
                )
            ]

        return targets

    async def fetch_connection_targets(
        self, connection: ConnRecord
    ) -> Sequence[ConnectionTarget]:
        """Get a list of connection targets from a `ConnRecord`.

        Args:
            connection: The connection record (with associated `DIDDoc`)
                used to generate the connection target
        """

        if not connection.my_did:
            self._logger.debug("No local DID associated with connection")
            return []

        async with self._profile.session() as session:
            wallet = session.inject(BaseWallet)
            my_info = await wallet.get_local_did(connection.my_did)

        if (
            ConnRecord.State.get(connection.state)
            in (ConnRecord.State.INVITATION, ConnRecord.State.REQUEST)
            and ConnRecord.Role.get(connection.their_role) is ConnRecord.Role.RESPONDER
        ):  # invitation received or sending request
            return await self._fetch_targets_for_connection_in_progress(
                connection, my_info.verkey
            )

        if not connection.their_did:
            self._logger.debug("No target DID associated with connection")
            return []

        return await self.resolve_connection_targets(
            connection.their_did, my_info.verkey, connection.their_label
        )

    async def get_connection_targets(
        self,
        *,
        connection_id: Optional[str] = None,
        connection: Optional[ConnRecord] = None,
    ):
        """Create a connection target from a `ConnRecord`.

        Args:
            connection_id: The connection ID to search for
            connection: The connection record itself, if already available
        """
        if connection_id is None and connection is None:
            raise ValueError("Must supply either connection_id or connection")

        if not connection_id:
            assert connection
            connection_id = connection.connection_id

        cache = self._profile.inject_or(BaseCache)
        cache_key = f"connection_target::{connection_id}"
        if cache:
            async with cache.acquire(cache_key) as entry:
                if entry.result:
                    self._logger.debug("Connection targets retrieved from cache")
                    targets = [ConnectionTarget.deserialize(row) for row in entry.result]
                else:
                    if not connection:
                        async with self._profile.session() as session:
                            connection = await ConnRecord.retrieve_by_id(
                                session, connection_id
                            )

                    targets = await self.fetch_connection_targets(connection)

                    if connection.state == ConnRecord.State.COMPLETED.rfc160:
                        # Only set cache if connection has reached completed state
                        # Otherwise, a replica that participated early in exchange
                        # may have bad data set in cache.
                        self._logger.debug("Caching connection targets")
                        await entry.set_result([row.serialize() for row in targets], 3600)
                    else:
                        self._logger.debug(
                            "Not caching connection targets for connection in state %s",
                            connection.state,
                        )
        else:
            if not connection:
                async with self._profile.session() as session:
                    connection = await ConnRecord.retrieve_by_id(session, connection_id)

            targets = await self.fetch_connection_targets(connection)
        return targets

    async def clear_connection_targets_cache(self, connection_id: str):
        """Clear the connection targets cache for a given connection ID.

        Historically, connections have not been updatable after the protocol
        completes. However, with DID Rotation, we need to be able to update
        the connection targets and clear the cache of targets.
        """
        # TODO it would be better to include the DIDs of the connection in the
        # target cache key This solution only works when using whole cluster
        # caching or have only a single instance with local caching
        cache = self._profile.inject_or(BaseCache)
        if cache:
            cache_key = f"connection_target::{connection_id}"
            await cache.clear(cache_key)

    def diddoc_connection_targets(
        self,
        doc: Optional[Union[DIDDoc, dict]],
        sender_verkey: str,
        their_label: Optional[str] = None,
    ) -> Sequence[ConnectionTarget]:
        """Get a list of connection targets from a DID Document.

        Args:
            doc: The DID Document to create the target from
            sender_verkey: The verkey we are using
            their_label: The connection label they are using
        """
        if isinstance(doc, dict):
            doc = DIDDoc.deserialize(doc)
        if not doc:
            raise BaseConnectionManagerError("No DIDDoc provided for connection target")
        if not doc.did:
            raise BaseConnectionManagerError("DIDDoc has no DID")
        if not doc.service:
            raise BaseConnectionManagerError("No services defined by DIDDoc")

        targets = []
        for service in doc.service.values():
            if service.recip_keys:
                targets.append(
                    ConnectionTarget(
                        did=doc.did,
                        endpoint=service.endpoint,
                        label=their_label,
                        recipient_keys=[key.value for key in (service.recip_keys or ())],
                        routing_keys=[key.value for key in (service.routing_keys or ())],
                        sender_key=sender_verkey,
                    )
                )
        return targets

    async def fetch_did_document(self, did: str) -> Tuple[dict, StorageRecord]:
        """Retrieve a DID Document for a given DID.

        Args:
            did: The DID to search for
        """
        async with self._profile.session() as session:
            storage = session.inject(BaseStorage)
            record = await storage.find_record(self.RECORD_TYPE_DID_DOC, {"did": did})
        return json.loads(record.value), record

    async def find_connection(
        self,
        their_did: Optional[str],
        my_did: Optional[str] = None,
        parent_thread_id: Optional[str] = None,
        auto_complete=False,
    ) -> Optional[ConnRecord]:
        """Look up existing connection information for a sender verkey.

        Args:
            their_did: Their DID
            my_did: My DID
            parent_thread_id: Parent thread ID
            auto_complete: Should this connection automatically be promoted to active

        Returns:
            The located `ConnRecord`, if any

        """
        connection = None
        if their_did and their_did.startswith("did:peer:4"):
            # did:peer:4 always recorded as long
            long = their_did
            short = self.long_did_peer_to_short(their_did)
            try:
                async with self._profile.session() as session:
                    connection = await ConnRecord.retrieve_by_did_peer_4(
                        session, long, short, my_did
                    )
            except StorageNotFoundError:
                pass
        elif their_did:
            try:
                async with self._profile.session() as session:
                    connection = await ConnRecord.retrieve_by_did(
                        session, their_did, my_did
                    )
            except StorageNotFoundError:
                pass

        if (
            connection
            and ConnRecord.State.get(connection.state) is ConnRecord.State.RESPONSE
            and auto_complete
        ):
            connection.state = ConnRecord.State.COMPLETED.rfc160
            async with self._profile.session() as session:
                await connection.save(session, reason="Connection promoted to active")
                if session.settings.get("auto_disclose_features"):
                    discovery_mgr = V20DiscoveryMgr(self._profile)
                    await discovery_mgr.proactive_disclose_features(
                        connection_id=connection.connection_id
                    )

        if not connection and parent_thread_id:
            async with self._profile.session() as session:
                connection = await ConnRecord.retrieve_by_invitation_msg_id(
                    session,
                    parent_thread_id,
                    their_role=ConnRecord.Role.REQUESTER.rfc160,
                )

        return connection

    async def find_inbound_connection(
        self, receipt: MessageReceipt
    ) -> Optional[ConnRecord]:
        """Deserialize an incoming message and further populate the request context.

        Args:
            receipt: The message receipt

        Returns:
            The `ConnRecord` associated with the expanded message, if any

        """

        cache_key = None
        connection = None
        resolved = False

        if receipt.sender_verkey and receipt.recipient_verkey:
            cache_key = (
                f"connection_by_verkey::{receipt.sender_verkey}"
                f"::{receipt.recipient_verkey}"
            )
            cache = self._profile.inject_or(BaseCache)
            if cache:
                async with cache.acquire(cache_key) as entry:
                    if entry.result:
                        cached = entry.result
                        receipt.sender_did = cached["sender_did"]
                        receipt.recipient_did_public = cached["recipient_did_public"]
                        receipt.recipient_did = cached["recipient_did"]
                        async with self._profile.session() as session:
                            connection = await ConnRecord.retrieve_by_id(
                                session, cached["id"]
                            )
                    else:
                        connection = await self.resolve_inbound_connection(receipt)
                        if connection:
                            cache_val = {
                                "id": connection.connection_id,
                                "sender_did": receipt.sender_did,
                                "recipient_did": receipt.recipient_did,
                                "recipient_did_public": receipt.recipient_did_public,
                            }
                            await entry.set_result(cache_val, 3600)
                        resolved = True

        if not connection and not resolved:
            connection = await self.resolve_inbound_connection(receipt)
        return connection

    async def resolve_inbound_connection(
        self, receipt: MessageReceipt
    ) -> Optional[ConnRecord]:
        """Populate the receipt DID information and find the related `ConnRecord`.

        Args:
            receipt: The message receipt

        Returns:
            The `ConnRecord` associated with the expanded message, if any

        """

        receipt.sender_did = None
        if receipt.sender_verkey:
            try:
                receipt.sender_did = await self.find_did_for_key(receipt.sender_verkey)
            except StorageNotFoundError:
                pass

        if receipt.recipient_verkey:
            try:
                async with self._profile.session() as session:
                    wallet = session.inject(BaseWallet)
                    my_info = await wallet.get_local_did_for_verkey(
                        receipt.recipient_verkey
                    )
                receipt.recipient_did = my_info.did
                if "posted" in my_info.metadata and my_info.metadata["posted"] is True:
                    receipt.recipient_did_public = True
            except InjectionError:
                self._logger.warning(
                    "Cannot resolve recipient verkey, no wallet defined by context: %s",
                    receipt.recipient_verkey,
                )
            except WalletNotFoundError:
                self._logger.debug(
                    "No corresponding DID found for recipient verkey: %s",
                    receipt.recipient_verkey,
                )

        return await self.find_connection(
            receipt.sender_did, receipt.recipient_did, receipt.parent_thread_id, True
        )

    async def get_endpoints(self, conn_id: str) -> Tuple[Optional[str], Optional[str]]:
        """Get connection endpoints.

        Args:
            conn_id: connection identifier

        Returns:
            Their endpoint for this connection

        """
        async with self._profile.session() as session:
            connection = await ConnRecord.retrieve_by_id(session, conn_id)
            wallet = session.inject(BaseWallet)
            my_did_info = await wallet.get_local_did(connection.my_did)

        my_endpoint = my_did_info.metadata.get(
            "endpoint",
            self._profile.settings.get("default_endpoint"),
        )

        conn_targets = await self.get_connection_targets(
            connection_id=connection.connection_id,
            connection=connection,
        )
        return (my_endpoint, conn_targets[0].endpoint)

    async def create_static_connection(
        self,
        my_did: Optional[str] = None,
        my_seed: Optional[str] = None,
        their_did: Optional[str] = None,
        their_seed: Optional[str] = None,
        their_verkey: Optional[str] = None,
        their_endpoint: Optional[str] = None,
        their_label: Optional[str] = None,
        alias: Optional[str] = None,
        mediation_id: Optional[str] = None,
    ) -> Tuple[DIDInfo, DIDInfo, ConnRecord]:
        """Register a new static connection (for use by the test suite).

        This method is used to create a new static connection. It allows overriding
        various parameters such as DIDs, seeds, verkeys, endpoints, labels, and aliases.
        It returns the created connection record along with the associated DID
        information.

        Args:
            my_did (Optional[str]): Override the DID used in the connection.
            my_seed (Optional[str]): Provide a seed used to generate our DID and keys.
            their_did (Optional[str]): Provide the DID used by the other party.
            their_seed (Optional[str]): Provide a seed used to generate their DID and
                keys.
            their_verkey (Optional[str]): Provide the verkey used by the other party.
            their_endpoint (Optional[str]): Their URL endpoint for routing messages.
            their_label (Optional[str]): An alias for this connection record.
            alias (Optional[str]): An alias for this connection record.
            mediation_id (Optional[str]): The mediation ID for routing through a mediator.

        Returns:
            Tuple[DIDInfo, DIDInfo, ConnRecord]: A tuple containing the following:
                - my DIDInfo: The DID information for the local party.
                - their DIDInfo: The DID information for the other party.
                - new `ConnRecord` instance: The newly created connection record.

        Raises:
            BaseConnectionManagerError: If either a verkey or seed must be provided for
                the other party.

        """
        async with self._profile.session() as session:
            wallet = session.inject(BaseWallet)
            # seed and DID optional
            my_info = await wallet.create_local_did(SOV, ED25519, my_seed, my_did)

        # must provide their DID and verkey if the seed is not known
        if (not their_did or not their_verkey) and not their_seed:
            raise BaseConnectionManagerError(
                "Either a verkey or seed must be provided for the other party"
            )
        if not their_did:
            their_did = seed_to_did(their_seed)
        if not their_verkey:
            their_verkey_bin, _ = create_keypair(ED25519, their_seed.encode())
            their_verkey = bytes_to_b58(their_verkey_bin)
        their_info = DIDInfo(their_did, their_verkey, {}, method=SOV, key_type=ED25519)

        # Create connection record
        connection = ConnRecord(
            invitation_mode=ConnRecord.INVITATION_MODE_STATIC,
            my_did=my_info.did,
            their_did=their_info.did,
            their_label=their_label,
            state=ConnRecord.State.COMPLETED.rfc160,
            alias=alias,
            connection_protocol=CONN_PROTO,
        )
        async with self._profile.session() as session:
            await connection.save(session, reason="Created new static connection")
            if session.settings.get("auto_disclose_features"):
                discovery_mgr = V20DiscoveryMgr(self._profile)
                await discovery_mgr.proactive_disclose_features(
                    connection_id=connection.connection_id
                )

        # Routing
        mediation_record = await self._route_manager.mediation_record_if_id(
            self._profile, mediation_id, or_default=True
        )

        multitenant_mgr = self._profile.inject_or(BaseMultitenantManager)
        wallet_id = self._profile.settings.get("wallet.id")

        base_mediation_record = None
        if multitenant_mgr and wallet_id:
            base_mediation_record = await multitenant_mgr.get_default_mediator()

        await self._route_manager.route_static(
            self._profile, connection, mediation_record
        )

        # Synthesize their DID doc
        did_doc = await self.create_did_document(
            their_info,
            [their_endpoint or ""],
            mediation_records=list(
                filter(None, [base_mediation_record, mediation_record])
            ),
        )

        await self.store_did_document(did_doc)

        return my_info, their_info, connection



"""Record used to represent a parameter in a menu form."""

from typing import Optional

from marshmallow import EXCLUDE, fields

from .....messaging.models.base import BaseModel, BaseModelSchema


class MenuFormParam(BaseModel):
    """Instance of a menu form param associated with an action menu option."""

    class Meta:
        """Menu form param metadata."""

        schema_class = "MenuFormParamSchema"

    def __init__(
        self,
        *,
        name: Optional[str] = None,
        title: Optional[str] = None,
        default: Optional[str] = None,
        description: Optional[str] = None,
        input_type: Optional[str] = None,
        required: Optional[bool] = None,
    ):
        """Initialize a MenuFormParam instance.

        Args:
            name: The parameter name
            title: The parameter title
            default: A default value for the parameter
            description: Additional descriptive text for the menu form parameter
            input_type: Input type
            required: Whether the parameter is required

        """
        self.name = name
        self.title = title
        self.default = default
        self.description = description
        self.input_type = input_type
        self.required = required


class MenuFormParamSchema(BaseModelSchema):
    """MenuFormParam schema."""

    class Meta:
        """MenuFormParamSchema metadata."""

        model_class = MenuFormParam
        unknown = EXCLUDE

    name = fields.Str(
        required=True,
        metadata={"description": "Menu parameter name", "example": "delay"},
    )
    title = fields.Str(
        required=True,
        metadata={"description": "Menu parameter title", "example": "Delay in seconds"},
    )
    default = fields.Str(
        required=False,
        metadata={"description": "Default parameter value", "example": "0"},
    )
    description = fields.Str(
        required=False,
        metadata={
            "description": "Additional descriptive text for menu form parameter",
            "example": "Delay in seconds before starting",
        },
    )
    input_type = fields.Str(
        required=False,
        data_key="type",
        metadata={"description": "Menu form parameter input type", "example": "int"},
    )
    required = fields.Bool(
        required=False,
        metadata={"description": "Whether parameter is required", "example": "False"},
    )



from .did_example_489398593 import DID_EXAMPLE_48939859
from .did_sov_QqEfJxe752NCmWqR5TssZ5 import DID_SOV_QqEfJxe752NCmWqR5TssZ5
from .did_z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL import (
    DID_z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL,
)
from .did_zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq import (
    DID_zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq,
)
from .did_zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa import (
    DID_zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa,
)

__all__ = [
    "DID_z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
    "DID_zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq",
    "DID_zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa",
    "DID_EXAMPLE_48939859",
    "DID_SOV_QqEfJxe752NCmWqR5TssZ5",
]



SECURITY_V1 = {
    "@context": {
        "id": "@id",
        "type": "@type",
        "dc": "http://purl.org/dc/terms/",
        "sec": "https://w3id.org/security#",
        "xsd": "http://www.w3.org/2001/XMLSchema#",
        "EcdsaKoblitzSignature2016": "sec:EcdsaKoblitzSignature2016",
        "Ed25519Signature2018": "sec:Ed25519Signature2018",
        "EncryptedMessage": "sec:EncryptedMessage",
        "GraphSignature2012": "sec:GraphSignature2012",
        "LinkedDataSignature2015": "sec:LinkedDataSignature2015",
        "LinkedDataSignature2016": "sec:LinkedDataSignature2016",
        "CryptographicKey": "sec:Key",
        "authenticationTag": "sec:authenticationTag",
        "canonicalizationAlgorithm": "sec:canonicalizationAlgorithm",
        "cipherAlgorithm": "sec:cipherAlgorithm",
        "cipherData": "sec:cipherData",
        "cipherKey": "sec:cipherKey",
        "created": {"@id": "dc:created", "@type": "xsd:dateTime"},
        "creator": {"@id": "dc:creator", "@type": "@id"},
        "digestAlgorithm": "sec:digestAlgorithm",
        "digestValue": "sec:digestValue",
        "domain": "sec:domain",
        "encryptionKey": "sec:encryptionKey",
        "expiration": {"@id": "sec:expiration", "@type": "xsd:dateTime"},
        "expires": {"@id": "sec:expiration", "@type": "xsd:dateTime"},
        "initializationVector": "sec:initializationVector",
        "iterationCount": "sec:iterationCount",
        "nonce": "sec:nonce",
        "normalizationAlgorithm": "sec:normalizationAlgorithm",
        "owner": {"@id": "sec:owner", "@type": "@id"},
        "password": "sec:password",
        "privateKey": {"@id": "sec:privateKey", "@type": "@id"},
        "privateKeyPem": "sec:privateKeyPem",
        "publicKey": {"@id": "sec:publicKey", "@type": "@id"},
        "publicKeyBase58": "sec:publicKeyBase58",
        "publicKeyPem": "sec:publicKeyPem",
        "publicKeyWif": "sec:publicKeyWif",
        "publicKeyService": {"@id": "sec:publicKeyService", "@type": "@id"},
        "revoked": {"@id": "sec:revoked", "@type": "xsd:dateTime"},
        "salt": "sec:salt",
        "signature": "sec:signature",
        "signatureAlgorithm": "sec:signingAlgorithm",
        "signatureValue": "sec:signatureValue",
    }
}



from typing import Coroutine, Optional, Union

from ....connections.models.conn_record import ConnRecord
from ....core.error import BaseError
from ....messaging.models.base_record import BaseRecord
from .message import ProblemReport


async def internal_error(
    err: BaseError,
    http_error_class,
    record: Union[ConnRecord, BaseRecord],
    outbound_handler: Coroutine,
    code: Optional[str] = None,
):
    """Send problem report and raise corresponding HTTP error."""
    if record:
        error_result = ProblemReport(
            description={"en": err.roll_up, "code": code or "abandoned"}
        )
        thid = getattr(record, "thread_id", None)
        if thid:
            error_result.assign_thread_id(thid)
        await outbound_handler(error_result, connection_id=record.connection_id)

    raise http_error_class(reason=err.roll_up) from err
